"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import json

import salt.modules.nftables as nftables
import salt.utils.files
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class NftablesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.nftables
    """

    def setup_loader_modules(self):
        return {nftables: {}}

    # 'version' function tests: 1

    def test_version(self):
        """
        Test if it return version from nftables --version
        """
        mock = MagicMock(return_value="nf_tables 0.3-1")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.version(), "0.3-1")

    # 'build_rule' function tests: 1

    def test_build_rule(self):
        """
        Test if it build a well-formatted nftables rule based on kwargs.
        """
        self.assertEqual(
            nftables.build_rule(full="True"),
            {"result": False, "rule": "", "comment": "Table needs to be specified"},
        )

        self.assertEqual(
            nftables.build_rule(table="filter", full="True"),
            {"result": False, "rule": "", "comment": "Chain needs to be specified"},
        )

        self.assertEqual(
            nftables.build_rule(table="filter", chain="input", full="True"),
            {"result": False, "rule": "", "comment": "Command needs to be specified"},
        )

        self.assertEqual(
            nftables.build_rule(
                table="filter",
                chain="input",
                command="insert",
                position="3",
                full="True",
            ),
            {
                "result": True,
                "rule": "nft insert rule ip filter input position 3 ",
                "comment": "Successfully built rule",
            },
        )

        self.assertEqual(
            nftables.build_rule(
                table="filter", chain="input", command="insert", full="True"
            ),
            {
                "result": True,
                "rule": "nft insert rule ip filter input ",
                "comment": "Successfully built rule",
            },
        )

        self.assertEqual(
            nftables.build_rule(
                table="filter", chain="input", command="halt", full="True"
            ),
            {
                "result": True,
                "rule": "nft halt rule ip filter input ",
                "comment": "Successfully built rule",
            },
        )

        self.assertEqual(
            nftables.build_rule(
                table="filter",
                chain="input",
                command="insert",
                position="3",
                full="True",
                connstate="related,established",
                saddr="10.0.0.1",
                daddr="10.0.0.2",
                jump="accept",
            ),
            {
                "result": True,
                "rule": (
                    "nft insert rule ip filter input position 3 ct state {"
                    " related,established } ip saddr 10.0.0.1 ip daddr 10.0.0.2 accept"
                ),
                "comment": "Successfully built rule",
            },
        )

        self.assertEqual(
            nftables.build_rule(), {"result": True, "rule": "", "comment": ""}
        )

    # 'get_saved_rules' function tests: 1

    def test_get_saved_rules(self):
        """
        Test if it return a data structure of the rules in the conf file
        """
        with patch.dict(nftables.__grains__, {"os_family": "Debian"}):
            with patch.object(salt.utils.files, "fopen", MagicMock(mock_open())):
                self.assertListEqual(nftables.get_saved_rules(), [])

    # 'list_tables' function tests: 1

    def test_list_tables(self):
        """
        Test if it return a data structure of the current, in-memory tables
        """
        list_tables = [{"family": "inet", "name": "filter", "handle": 2}]
        list_tables_mock = MagicMock(return_value=list_tables)

        with patch.object(nftables, "list_tables", list_tables_mock):
            self.assertListEqual(nftables.list_tables(), list_tables)

        list_tables_mock = MagicMock(return_value=[])
        with patch.object(nftables, "list_tables", list_tables_mock):
            self.assertListEqual(nftables.list_tables(), [])

    # 'get_rules' function tests: 1

    def test_get_rules(self):
        """
        Test if it return a data structure of the current, in-memory rules
        """
        list_tables_mock = MagicMock(
            return_value=[{"family": "inet", "name": "filter", "handle": 2}]
        )
        list_rules_return = """table inet filter {
            chain input {
                type filter hook input priority 0; policy accept;
            }

            chain forward {
                type filter hook forward priority 0; policy accept;
            }

            chain output {
                type filter hook output priority 0; policy accept;
            }
        }"""
        list_rules_mock = MagicMock(return_value=list_rules_return)
        expected = [list_rules_return]

        with patch.object(nftables, "list_tables", list_tables_mock):
            with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
                self.assertListEqual(nftables.get_rules(), expected)

        list_tables_mock = MagicMock(return_value=[])
        with patch.object(nftables, "list_tables", list_tables_mock):
            self.assertListEqual(nftables.get_rules(), [])

    # 'get_rules_json' function tests: 1

    def test_get_rules_json(self):
        """
        Test if it return a data structure of the current, in-memory rules
        """
        list_rules_return = """
        {
          "nftables": [
            {
              "table": {
                "family": "ip",
                "name": "filter",
                "handle": 47
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "input",
                "handle": 1,
                "type": "filter",
                "hook": "input",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "forward",
                "handle": 2,
                "type": "filter",
                "hook": "forward",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "output",
                "handle": 3,
                "type": "filter",
                "hook": "output",
                "prio": 0,
                "policy": "accept"
              }
            }
          ]
        }
        """
        list_rules_mock = MagicMock(return_value=list_rules_return)
        expected = json.loads(list_rules_return)["nftables"]

        with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
            self.assertListEqual(nftables.get_rules_json(), expected)

        list_rules_mock = MagicMock(return_value=[])
        with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
            self.assertListEqual(nftables.get_rules_json(), [])

    # 'save' function tests: 1

    def test_save(self):
        """
        Test if it save the current in-memory rules to disk
        """
        with patch.dict(nftables.__grains__, {"os_family": "Debian"}):
            mock = MagicMock(return_value=False)
            with patch.dict(nftables.__salt__, {"file.directory_exists": mock}):
                with patch.dict(nftables.__salt__, {"cmd.run": mock}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        self.assertEqual(nftables.save(), "#! nft -f\n\n")

                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(side_effect=IOError)
                    ):
                        self.assertRaises(CommandExecutionError, nftables.save)

    # 'get_rule_handle' function tests: 1

    def test_get_rule_handle(self):
        """
        Test if it get the handle for a particular rule
        """
        self.assertEqual(
            nftables.get_rule_handle(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        self.assertEqual(
            nftables.get_rule_handle(chain="input"),
            {"result": False, "comment": "Rule needs to be specified"},
        )

        _ru = "input tcp dport 22 log accept"
        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.get_rule_handle(chain="input", rule=_ru), ret)

        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 does not exist",
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.get_rule_handle(chain="input", rule=_ru), ret)

        ret = {
            "result": False,
            "comment": (
                "Rule input tcp dport 22 log accept chain"
                " input in table filter in family ipv4 does not exist"
            ),
        }
        ret1 = {
            "result": False,
            "comment": "Could not find rule input tcp dport 22 log accept",
        }
        with patch.object(
            nftables,
            "check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            with patch.object(
                nftables,
                "check_chain",
                MagicMock(return_value={"result": True, "comment": ""}),
            ):
                _ret1 = {
                    "result": False,
                    "comment": (
                        "Rule input tcp dport 22 log accept"
                        " chain input in table filter in"
                        " family ipv4 does not exist"
                    ),
                }
                _ret2 = {"result": True, "comment": ""}
                with patch.object(
                    nftables, "check", MagicMock(side_effect=[_ret1, _ret2])
                ):
                    self.assertEqual(
                        nftables.get_rule_handle(chain="input", rule=_ru), ret
                    )

                    _ru = "input tcp dport 22 log accept"
                    mock = MagicMock(return_value="")
                    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
                        self.assertEqual(
                            nftables.get_rule_handle(chain="input", rule=_ru), ret1
                        )

    # 'check' function tests: 1

    def test_check(self):
        """
        Test if it check for the existence of a rule in the table and chain
        """
        self.assertEqual(
            nftables.check(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        self.assertEqual(
            nftables.check(chain="input"),
            {"result": False, "comment": "Rule needs to be specified"},
        )

        _ru = "tcp dport 22 log accept"
        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check(chain="input", rule=_ru), ret)

        mock = MagicMock(return_value="table ip filter")
        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 does not exist",
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check(chain="input", rule=_ru), ret)

        mock = MagicMock(return_value="table ip filter chain input {{")
        ret = {
            "result": False,
            "comment": (
                "Rule tcp dport 22 log accept in chain input in table filter in family"
                " ipv4 does not exist"
            ),
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check(chain="input", rule=_ru), ret)

        r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
        mock = MagicMock(return_value=r_val)
        ret = {
            "result": True,
            "comment": (
                "Rule tcp dport 22 log accept in chain input in table filter in family"
                " ipv4 exists"
            ),
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check(chain="input", rule=_ru), ret)

    # 'check_chain' function tests: 1

    def test_check_chain(self):
        """
        Test if it check for the existence of a chain in the table
        """
        self.assertEqual(
            nftables.check_chain(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        mock = MagicMock(return_value="")
        ret = {
            "comment": "Chain input in table filter in family ipv4 does not exist",
            "result": False,
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check_chain(chain="input"), ret)

        mock = MagicMock(return_value="chain input {{")
        ret = {
            "comment": "Chain input in table filter in family ipv4 exists",
            "result": True,
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check_chain(chain="input"), ret)

    # 'check_table' function tests: 1

    def test_check_table(self):
        """
        Test if it check for the existence of a table
        """
        self.assertEqual(
            nftables.check_table(),
            {"result": False, "comment": "Table needs to be specified"},
        )

        mock = MagicMock(return_value="")
        ret = {"comment": "Table nat in family ipv4 does not exist", "result": False}
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check_table(table="nat"), ret)

        mock = MagicMock(return_value="table ip nat")
        ret = {"comment": "Table nat in family ipv4 exists", "result": True}
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.check_table(table="nat"), ret)

    # 'new_table' function tests: 1

    def test_new_table(self):
        """
        Test if it create new custom table.
        """
        self.assertEqual(
            nftables.new_table(table=None),
            {"result": False, "comment": "Table needs to be specified"},
        )

        mock = MagicMock(return_value="")
        ret = {"comment": "Table nat in family ipv4 created", "result": True}
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.new_table(table="nat"), ret)

        mock = MagicMock(return_value="table ip nat")
        ret = {"comment": "Table nat in family ipv4 exists", "result": True}
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.new_table(table="nat"), ret)

    # 'delete_table' function tests: 1

    def test_delete_table(self):
        """
        Test if it delete custom table.
        """
        self.assertEqual(
            nftables.delete_table(table=None),
            {"result": False, "comment": "Table needs to be specified"},
        )

        mock_ret = {
            "result": False,
            "comment": "Table nat in family ipv4 does not exist",
        }
        with patch(
            "salt.modules.nftables.check_table", MagicMock(return_value=mock_ret)
        ):
            ret = nftables.delete_table(table="nat")
            self.assertEqual(
                ret,
                {"result": False, "comment": "Table nat in family ipv4 does not exist"},
            )

        mock = MagicMock(return_value="table ip nat")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            self.assertEqual(
                nftables.delete_table(table="nat"),
                {
                    "comment": "Table nat in family ipv4 could not be deleted",
                    "result": False,
                },
            )

        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            self.assertEqual(
                nftables.delete_table(table="nat"),
                {"comment": "Table nat in family ipv4 deleted", "result": True},
            )

    # 'new_chain' function tests: 2

    def test_new_chain(self):
        """
        Test if it create new chain to the specified table.
        """
        self.assertEqual(
            nftables.new_chain(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.new_chain(chain="input"), ret)

        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 already exists",
        }
        mock = MagicMock(return_value="table ip filter chain input {{")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.new_chain(chain="input"), ret)

    def test_new_chain_variable(self):
        """
        Test if it create new chain to the specified table.
        """
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": False, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            self.assertEqual(
                nftables.new_chain(chain="input", table_type="filter"),
                {
                    "result": False,
                    "comment": "Table_type, hook, and priority required.",
                },
            )

            self.assertTrue(
                nftables.new_chain(
                    chain="input", table_type="filter", hook="input", priority=0
                )
            )

    # 'delete_chain' function tests: 1

    def test_delete_chain(self):
        """
        Test if it delete the chain from the specified table.
        """
        self.assertEqual(
            nftables.delete_chain(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.delete_chain(chain="input"), ret)

        ret = {
            "result": False,
            "comment": (
                "Chain input in table filter in family ipv4 could not be deleted"
            ),
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            self.assertEqual(nftables.delete_chain(chain="input"), ret)

        ret = {
            "result": True,
            "comment": "Chain input in table filter in family ipv4 deleted",
        }
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            self.assertEqual(nftables.delete_chain(chain="input"), ret)

    def test_delete_chain_variables(self):
        """
        Test if it delete the chain from the specified table.
        """
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _expected = {
                "comment": "Chain input in table filter in family ipv4 deleted",
                "result": True,
            }
            self.assertEqual(nftables.delete_chain(chain="input"), _expected)

    # 'append' function tests: 2

    def test_append(self):
        """
        Test if it append a rule to the specified table & chain.
        """
        self.assertEqual(
            nftables.append(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        self.assertEqual(
            nftables.append(chain="input"),
            {"result": False, "comment": "Rule needs to be specified"},
        )

        _ru = "input tcp dport 22 log accept"
        ret = {"comment": "Table filter in family ipv4 does not exist", "result": False}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.append(chain="input", rule=_ru), ret)

        ret = {
            "comment": "Chain input in table filter in family ipv4 does not exist",
            "result": False,
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.append(chain="input", rule=_ru), ret)

        r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
        mock = MagicMock(return_value=r_val)
        _expected = {
            "comment": (
                "Rule input tcp dport 22 log accept chain input in table filter in"
                " family ipv4 already exists"
            ),
            "result": False,
        }
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.append(chain="input", rule=_ru), _expected)

    def test_append_rule(self):
        """
        Test if it append a rule to the specified table & chain.
        """
        _ru = "input tcp dport 22 log accept"
        mock = MagicMock(side_effect=["1", ""])
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check",
            MagicMock(return_value={"result": False, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _expected = {
                "comment": (
                    'Failed to add rule "{}" chain input in table filter in family'
                    " ipv4.".format(_ru)
                ),
                "result": False,
            }
            self.assertEqual(nftables.append(chain="input", rule=_ru), _expected)
            _expected = {
                "comment": (
                    'Added rule "{}" chain input in table filter in family ipv4.'.format(
                        _ru
                    )
                ),
                "result": True,
            }
            self.assertEqual(nftables.append(chain="input", rule=_ru), _expected)

    # 'insert' function tests: 2

    def test_insert(self):
        """
        Test if it insert a rule into the specified table & chain,
        at the specified position.
        """
        self.assertEqual(
            nftables.insert(),
            {"result": False, "comment": "Chain needs to be specified"},
        )

        self.assertEqual(
            nftables.insert(chain="input"),
            {"result": False, "comment": "Rule needs to be specified"},
        )

        _ru = "input tcp dport 22 log accept"
        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.insert(chain="input", rule=_ru), ret)

        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 does not exist",
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.insert(chain="input", rule=_ru), ret)

        r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
        mock = MagicMock(return_value=r_val)
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            res = nftables.insert(chain="input", rule=_ru)
            import logging

            log = logging.getLogger(__name__)
            log.debug("=== res %s ===", res)
            self.assertTrue(nftables.insert(chain="input", rule=_ru))

    def test_insert_rule(self):
        """
        Test if it insert a rule into the specified table & chain,
        at the specified position.
        """
        _ru = "input tcp dport 22 log accept"
        mock = MagicMock(side_effect=["1", ""])
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check",
            MagicMock(return_value={"result": False, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _expected = {
                "result": False,
                "comment": (
                    'Failed to add rule "{}" chain input in table filter in family'
                    " ipv4.".format(_ru)
                ),
            }
            self.assertEqual(nftables.insert(chain="input", rule=_ru), _expected)
            _expected = {
                "result": True,
                "comment": (
                    'Added rule "{}" chain input in table filter in family ipv4.'.format(
                        _ru
                    )
                ),
            }
            self.assertEqual(nftables.insert(chain="input", rule=_ru), _expected)

    # 'delete' function tests: 2

    def test_delete(self):
        """
        Test if it delete a rule from the specified table & chain,
        specifying either the rule in its entirety, or
        the rule's position in the chain.
        """
        _ru = "input tcp dport 22 log accept"
        ret = {
            "result": False,
            "comment": "Only specify a position or a rule, not both",
        }
        self.assertEqual(
            nftables.delete(table="filter", chain="input", position="3", rule=_ru), ret
        )

        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(
                nftables.delete(table="filter", chain="input", rule=_ru), ret
            )

        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 does not exist",
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(
                nftables.delete(table="filter", chain="input", rule=_ru), ret
            )

        mock = MagicMock(return_value="table ip filter chain input {{")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertTrue(nftables.delete(table="filter", chain="input", rule=_ru))

    def test_delete_rule(self):
        """
        Test if it delete a rule from the specified table & chain,
        specifying either the rule in its entirety, or
        the rule's position in the chain.
        """
        mock = MagicMock(side_effect=["1", ""])
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _expected = {
                "result": False,
                "comment": (
                    'Failed to delete rule "None" in chain input  table filter in'
                    " family ipv4"
                ),
            }
            self.assertEqual(
                nftables.delete(table="filter", chain="input", position="3"), _expected
            )
            _expected = {
                "result": True,
                "comment": (
                    'Deleted rule "None" in chain input in table filter in family ipv4.'
                ),
            }
            self.assertEqual(
                nftables.delete(table="filter", chain="input", position="3"), _expected
            )

    # 'flush' function tests: 2

    def test_flush(self):
        """
        Test if it flush the chain in the specified table, flush all chains
        in the specified table if chain is not specified.
        """
        ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.flush(table="filter", chain="input"), ret)

        ret = {
            "result": False,
            "comment": "Chain input in table filter in family ipv4 does not exist",
        }
        mock = MagicMock(return_value="table ip filter")
        with patch.dict(nftables.__salt__, {"cmd.run": mock}):
            self.assertEqual(nftables.flush(table="filter", chain="input"), ret)

    def test_flush_chain(self):
        """
        Test if it flush the chain in the specified table, flush all chains
        in the specified table if chain is not specified.
        """
        mock = MagicMock(side_effect=["1", ""])
        with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.nftables.check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ), patch(
            "salt.modules.nftables.check_table",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _expected = {
                "result": False,
                "comment": (
                    "Failed to flush rules from chain input in table filter in family"
                    " ipv4."
                ),
            }
            self.assertEqual(nftables.flush(table="filter", chain="input"), _expected)
            _expected = {
                "result": True,
                "comment": (
                    "Flushed rules from chain input in table filter in family ipv4."
                ),
            }
            self.assertEqual(nftables.flush(table="filter", chain="input"), _expected)

    # 'get_policy' function tests: 1

    def test_get_policy(self):
        """
        Test the current policy for the specified table/chain
        """
        list_rules_return = """
        {
          "nftables": [
            {
              "table": {
                "family": "ip",
                "name": "filter",
                "handle": 47
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "input",
                "handle": 1,
                "type": "filter",
                "hook": "input",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "forward",
                "handle": 2,
                "type": "filter",
                "hook": "forward",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "output",
                "handle": 3,
                "type": "filter",
                "hook": "output",
                "prio": 0,
                "policy": "accept"
              }
            }
          ]
        }
        """
        expected = json.loads(list_rules_return)

        self.assertEqual(
            nftables.get_policy(table="filter", chain=None, family="ipv4"),
            "Error: Chain needs to be specified",
        )

        with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
            self.assertEqual(
                nftables.get_policy(table="filter", chain="input", family="ipv4"),
                "accept",
            )

        with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
            self.assertIsNone(
                nftables.get_policy(table="filter", chain="missing", family="ipv4")
            )

    # 'set_policy' function tests: 1

    def test_set_policy(self):
        """
        Test set the current policy for the specified table/chain
        """
        list_rules_return = """
        {
          "nftables": [
            {
              "table": {
                "family": "ip",
                "name": "filter",
                "handle": 47
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "input",
                "handle": 1,
                "type": "filter",
                "hook": "input",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "forward",
                "handle": 2,
                "type": "filter",
                "hook": "forward",
                "prio": 0,
                "policy": "accept"
              }
            },
            {
              "chain": {
                "family": "ip",
                "table": "filter",
                "name": "output",
                "handle": 3,
                "type": "filter",
                "hook": "output",
                "prio": 0,
                "policy": "accept"
              }
            }
          ]
        }
        """
        expected = json.loads(list_rules_return)["nftables"]

        self.assertEqual(
            nftables.set_policy(table="filter", chain=None, policy=None, family="ipv4"),
            "Error: Chain needs to be specified",
        )

        self.assertEqual(
            nftables.set_policy(
                table="filter", chain="input", policy=None, family="ipv4"
            ),
            "Error: Policy needs to be specified",
        )

        mock = MagicMock(return_value={"retcode": 0})
        with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
            with patch.dict(nftables.__salt__, {"cmd.run_all": mock}):
                self.assertTrue(
                    nftables.set_policy(
                        table="filter", chain="input", policy="accept", family="ipv4"
                    )
                )



"""
Unit Tests for functions located in salt.utils.state.py.
"""


import copy
import textwrap

import salt.utils.odict
import salt.utils.state
from tests.support.unit import TestCase


class StateUtilTestCase(TestCase):
    """
    Test case for state util.
    """

    def test_check_result(self):
        self.assertFalse(
            salt.utils.state.check_result(None),
            "Failed to handle None as an invalid data type.",
        )
        self.assertFalse(
            salt.utils.state.check_result([]), "Failed to handle an invalid data type."
        )
        self.assertFalse(
            salt.utils.state.check_result({}), "Failed to handle an empty dictionary."
        )
        self.assertFalse(
            salt.utils.state.check_result({"host1": []}),
            "Failed to handle an invalid host data structure.",
        )
        test_valid_state = {"host1": {"test_state": {"result": "We have liftoff!"}}}
        self.assertTrue(salt.utils.state.check_result(test_valid_state))
        test_valid_false_states = {
            "test1": salt.utils.odict.OrderedDict(
                [
                    (
                        "host1",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": False}),
                            ]
                        ),
                    ),
                ]
            ),
            "test2": salt.utils.odict.OrderedDict(
                [
                    (
                        "host1",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": True}),
                            ]
                        ),
                    ),
                    (
                        "host2",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": False}),
                            ]
                        ),
                    ),
                ]
            ),
            "test3": ["a"],
            "test4": salt.utils.odict.OrderedDict(
                [
                    (
                        "asup",
                        salt.utils.odict.OrderedDict(
                            [
                                (
                                    "host1",
                                    salt.utils.odict.OrderedDict(
                                        [
                                            ("test_state0", {"result": True}),
                                            ("test_state", {"result": True}),
                                        ]
                                    ),
                                ),
                                (
                                    "host2",
                                    salt.utils.odict.OrderedDict(
                                        [
                                            ("test_state0", {"result": True}),
                                            ("test_state", {"result": False}),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    )
                ]
            ),
            "test5": salt.utils.odict.OrderedDict(
                [
                    (
                        "asup",
                        salt.utils.odict.OrderedDict(
                            [
                                (
                                    "host1",
                                    salt.utils.odict.OrderedDict(
                                        [
                                            ("test_state0", {"result": True}),
                                            ("test_state", {"result": True}),
                                        ]
                                    ),
                                ),
                                ("host2", salt.utils.odict.OrderedDict([])),
                            ]
                        ),
                    )
                ]
            ),
        }
        for test, data in test_valid_false_states.items():
            self.assertFalse(
                salt.utils.state.check_result(data), msg="{} failed".format(test)
            )
        test_valid_true_states = {
            "test1": salt.utils.odict.OrderedDict(
                [
                    (
                        "host1",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": True}),
                            ]
                        ),
                    ),
                ]
            ),
            "test3": salt.utils.odict.OrderedDict(
                [
                    (
                        "host1",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": True}),
                            ]
                        ),
                    ),
                    (
                        "host2",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": True}),
                            ]
                        ),
                    ),
                ]
            ),
            "test4": salt.utils.odict.OrderedDict(
                [
                    (
                        "asup",
                        salt.utils.odict.OrderedDict(
                            [
                                (
                                    "host1",
                                    salt.utils.odict.OrderedDict(
                                        [
                                            ("test_state0", {"result": True}),
                                            ("test_state", {"result": True}),
                                        ]
                                    ),
                                ),
                                (
                                    "host2",
                                    salt.utils.odict.OrderedDict(
                                        [
                                            ("test_state0", {"result": True}),
                                            ("test_state", {"result": True}),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    )
                ]
            ),
            "test2": salt.utils.odict.OrderedDict(
                [
                    (
                        "host1",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": None}),
                                ("test_state", {"result": True}),
                            ]
                        ),
                    ),
                    (
                        "host2",
                        salt.utils.odict.OrderedDict(
                            [
                                ("test_state0", {"result": True}),
                                ("test_state", {"result": "abc"}),
                            ]
                        ),
                    ),
                ]
            ),
        }
        for test, data in test_valid_true_states.items():
            self.assertTrue(
                salt.utils.state.check_result(data), msg="{} failed".format(test)
            )
        test_invalid_true_ht_states = {
            "test_onfail_simple2": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    ("test_vstate0", {"result": False}),
                                    ("test_vstate1", {"result": True}),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_vstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_vstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", True),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_vstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail_integ2": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "t_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate1_|-echo_|-run",
                                        {"result": False},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_ivstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                        "t": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_ivstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail_integ3": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "t_|-test_ivstate0_|-echo_|-run",
                                        {"result": True},
                                    ),
                                    (
                                        "cmd_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate1_|-echo_|-run",
                                        {"result": False},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_ivstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                        "t": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_ivstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail_integ4": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "t_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate1_|-echo_|-run",
                                        {"result": True},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_ivstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                        "t": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_ivstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                    "test_ivstate2": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", True),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    ("test_state0", {"result": False}),
                                    ("test_state", {"result": True}),
                                ]
                            ),
                        ),
                    ]
                ),
                None,
            ),
            "test_onfail_d": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    ("test_state0", {"result": False}),
                                    ("test_state", {"result": True}),
                                ]
                            ),
                        ),
                    ]
                ),
                {},
            ),
        }
        for test, testdata in test_invalid_true_ht_states.items():
            data, ht = testdata
            for t_ in [a for a in data["host1"]]:
                tdata = data["host1"][t_]
                if "_|-" in t_:
                    t_ = t_.split("_|-")[1]
                tdata["__id__"] = t_
            self.assertFalse(
                salt.utils.state.check_result(data, highstate=ht),
                msg="{} failed".format(test),
            )

        test_valid_true_ht_states = {
            "test_onfail_integ": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "cmd_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate1_|-echo_|-run",
                                        {"result": True},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_ivstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_ivstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail_intega3": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "t_|-test_ivstate0_|-echo_|-run",
                                        {"result": True},
                                    ),
                                    (
                                        "cmd_|-test_ivstate0_|-echo_|-run",
                                        {"result": False},
                                    ),
                                    (
                                        "cmd_|-test_ivstate1_|-echo_|-run",
                                        {"result": True},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_ivstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                        "t": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_ivstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_ivstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
            "test_onfail_simple": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    ("test_vstate0", {"result": False}),
                                    ("test_vstate1", {"result": True}),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_vstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_vstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    ("onfail_stop", False),
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_vstate0")]
                                            )
                                        ],
                                    ),
                                ]
                            ),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),  # order is different
            "test_onfail_simple_rev": (
                salt.utils.odict.OrderedDict(
                    [
                        (
                            "host1",
                            salt.utils.odict.OrderedDict(
                                [
                                    ("test_vstate0", {"result": False}),
                                    ("test_vstate1", {"result": True}),
                                ]
                            ),
                        ),
                    ]
                ),
                {
                    "test_vstate0": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            "run",
                            {"order": 10002},
                        ],
                    },
                    "test_vstate1": {
                        "__env__": "base",
                        "__sls__": "a",
                        "cmd": [
                            salt.utils.odict.OrderedDict([("name", "/bin/true")]),
                            salt.utils.odict.OrderedDict(
                                [
                                    (
                                        "onfail",
                                        [
                                            salt.utils.odict.OrderedDict(
                                                [("cmd", "test_vstate0")]
                                            )
                                        ],
                                    )
                                ]
                            ),
                            salt.utils.odict.OrderedDict([("onfail_stop", False)]),
                            "run",
                            {"order": 10004},
                        ],
                    },
                },
            ),
        }
        for test, testdata in test_valid_true_ht_states.items():
            data, ht = testdata
            for t_ in [a for a in data["host1"]]:
                tdata = data["host1"][t_]
                if "_|-" in t_:
                    t_ = t_.split("_|-")[1]
                tdata["__id__"] = t_
            self.assertTrue(
                salt.utils.state.check_result(data, highstate=ht),
                msg="{} failed".format(test),
            )
        test_valid_false_state = {"host1": {"test_state": {"result": False}}}
        self.assertFalse(salt.utils.state.check_result(test_valid_false_state))


class UtilStateMergeSubreturnTestcase(TestCase):
    """
    Test cases for salt.utils.state.merge_subreturn function.
    """

    main_ret = {
        "name": "primary",
        # result may be missing, as primarysalt.utils.state is still in progress
        "comment": "",
        "changes": {},
    }
    sub_ret = {
        "name": "secondary",
        "result": True,
        "comment": "",
        "changes": {},
    }

    def test_merge_result(self):
        # result not created if not needed
        for no_effect_result in [True, None]:
            m = copy.deepcopy(self.main_ret)
            s = copy.deepcopy(self.sub_ret)
            s["result"] = no_effect_result
            res = salt.utils.state.merge_subreturn(m, s)
            self.assertNotIn("result", res)

        # False subresult is propagated to existing result
        for original_result in [True, None, False]:
            m = copy.deepcopy(self.main_ret)
            m["result"] = original_result
            s = copy.deepcopy(self.sub_ret)
            s["result"] = False
            res = salt.utils.state.merge_subreturn(m, s)
            self.assertFalse(res["result"])

        # False result cannot be overridden
        for any_result in [True, None, False]:
            m = copy.deepcopy(self.main_ret)
            m["result"] = False
            s = copy.deepcopy(self.sub_ret)
            s["result"] = any_result
            res = salt.utils.state.merge_subreturn(m, s)
            self.assertFalse(res["result"])

    def test_merge_changes(self):
        # The main changes dict should always already exist,
        # and there should always be a changes dict in the secondary.
        primary_changes = {"old": None, "new": "my_resource"}
        secondary_changes = {"old": None, "new": ["alarm-1", "alarm-2"]}

        # No changes case
        m = copy.deepcopy(self.main_ret)
        s = copy.deepcopy(self.sub_ret)
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertDictEqual(res["changes"], {})

        # New changes don't get rid of existing changes
        m = copy.deepcopy(self.main_ret)
        m["changes"] = copy.deepcopy(primary_changes)
        s = copy.deepcopy(self.sub_ret)
        s["changes"] = copy.deepcopy(secondary_changes)
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertDictEqual(
            res["changes"],
            {"old": None, "new": "my_resource", "secondary": secondary_changes},
        )

        # The subkey parameter is respected
        m = copy.deepcopy(self.main_ret)
        m["changes"] = copy.deepcopy(primary_changes)
        s = copy.deepcopy(self.sub_ret)
        s["changes"] = copy.deepcopy(secondary_changes)
        res = salt.utils.state.merge_subreturn(m, s, subkey="alarms")
        self.assertDictEqual(
            res["changes"],
            {"old": None, "new": "my_resource", "alarms": secondary_changes},
        )

    def test_merge_comments(self):
        main_comment_1 = "First primary comment."
        main_comment_2 = "Second primary comment."
        sub_comment_1 = "First secondary comment,\nwhich spans two lines."
        sub_comment_2 = "Second secondary comment: {}".format(
            "some error\n  And a traceback",
        )
        final_comment = textwrap.dedent(
            """\
            First primary comment.
            Second primary comment.
            First secondary comment,
            which spans two lines.
            Second secondary comment: some error
              And a traceback
        """.rstrip()
        )

        # Joining two strings
        m = copy.deepcopy(self.main_ret)
        m["comment"] = main_comment_1 + "\n" + main_comment_2
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = sub_comment_1 + "\n" + sub_comment_2
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertMultiLineEqual(res["comment"], final_comment)

        # Joining string and a list
        m = copy.deepcopy(self.main_ret)
        m["comment"] = main_comment_1 + "\n" + main_comment_2
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = [sub_comment_1, sub_comment_2]
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertMultiLineEqual(res["comment"], final_comment)

        # For tests where output is a list,
        # also test that final joined output will match
        # Joining list and a string
        m = copy.deepcopy(self.main_ret)
        m["comment"] = [main_comment_1, main_comment_2]
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = sub_comment_1 + "\n" + sub_comment_2
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(
            res["comment"],
            [main_comment_1, main_comment_2, sub_comment_1 + "\n" + sub_comment_2],
        )
        self.assertMultiLineEqual("\n".join(res["comment"]), final_comment)

        # Joining two lists
        m = copy.deepcopy(self.main_ret)
        m["comment"] = [main_comment_1, main_comment_2]
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = [sub_comment_1, sub_comment_2]
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(
            res["comment"],
            [main_comment_1, main_comment_2, sub_comment_1, sub_comment_2],
        )
        self.assertMultiLineEqual("\n".join(res["comment"]), final_comment)

    def test_merge_empty_comments(self):
        # Since the primarysalt.utils.state is in progress,
        # the main comment may be empty, either '' or [].
        # Note that [''] is a degenerate case and should never happen,
        # hence the behavior is left unspecified in that case.
        # The secondary comment should never be empty,
        # because thatsalt.utils.state has already returned,
        # so we leave the behavior unspecified in that case.
        sub_comment_1 = "Secondary comment about changes:"
        sub_comment_2 = "A diff that goes with the previous comment"
        # No contributions from primary
        final_comment = sub_comment_1 + "\n" + sub_comment_2

        # Joining empty string and a string
        m = copy.deepcopy(self.main_ret)
        m["comment"] = ""
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = sub_comment_1 + "\n" + sub_comment_2
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(res["comment"], final_comment)

        # Joining empty string and a list
        m = copy.deepcopy(self.main_ret)
        m["comment"] = ""
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = [sub_comment_1, sub_comment_2]
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(res["comment"], final_comment)

        # For tests where output is a list,
        # also test that final joined output will match
        # Joining empty list and a string
        m = copy.deepcopy(self.main_ret)
        m["comment"] = []
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = sub_comment_1 + "\n" + sub_comment_2
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(res["comment"], [final_comment])
        self.assertEqual("\n".join(res["comment"]), final_comment)

        # Joining empty list and a list
        m = copy.deepcopy(self.main_ret)
        m["comment"] = []
        s = copy.deepcopy(self.sub_ret)
        s["comment"] = [sub_comment_1, sub_comment_2]
        res = salt.utils.state.merge_subreturn(m, s)
        self.assertEqual(res["comment"], [sub_comment_1, sub_comment_2])
        self.assertEqual("\n".join(res["comment"]), final_comment)



import functools
import logging
import sys
import time
import types

if sys.version_info >= (3, 10):
    # Python 3.10 will include a fix in importlib.metadata which allows us to
    # get the distribution of a loaded entry-point
    import importlib.metadata  # pylint: disable=no-member,no-name-in-module

    USE_IMPORTLIB_METADATA_STDLIB = True
else:
    USE_IMPORTLIB_METADATA_STDLIB = False
    try:
        from salt._compat import importlib_metadata

        USE_IMPORTLIB_METADATA = True
    except ImportError:
        USE_IMPORTLIB_METADATA = False

log = logging.getLogger(__name__)


def timed_lru_cache(timeout_seconds, *, maxsize=256, typed=False):
    """
    This decorator is the same in behavior as functools.lru_cache with the
    exception that it times out after the provided ``timeout_seconds``
    """

    def _wrapper(f):
        # Apply @lru_cache to f
        f = functools.lru_cache(maxsize=maxsize, typed=typed)(f)
        f.delta = timeout_seconds
        f.expiration = time.monotonic() + f.delta

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            now = time.monotonic()
            if now >= f.expiration:
                f.cache_clear()
                f.expiration = now + f.delta
            return f(*args, **kwargs)

        return _wrapped

    return _wrapper


@timed_lru_cache(timeout_seconds=0.5)
def iter_entry_points(group, name=None):
    entry_points_listing = []
    if USE_IMPORTLIB_METADATA_STDLIB:
        log.debug("Using importlib.metadata to load entry points")
        entry_points = importlib.metadata.entry_points()
    elif USE_IMPORTLIB_METADATA:
        log.debug("Using importlib_metadata to load entry points")
        entry_points = importlib_metadata.entry_points()
    else:
        return entry_points_listing

    for entry_point_group, entry_points_list in entry_points.items():
        if entry_point_group != group:
            continue
        for entry_point in entry_points_list:
            if name is not None and entry_point.name != name:
                continue
            entry_points_listing.append(entry_point)

    return entry_points_listing


def name_and_version_from_entry_point(entry_point):
    return types.SimpleNamespace(
        name=entry_point.dist.metadata["name"],
        version=entry_point.dist.version,
    )



import pytest
import salt.states.virt as virt
from tests.support.mock import MagicMock


class LibvirtMock(MagicMock):  # pylint: disable=too-many-ancestors
    """
    Libvirt library mock
    """

    class libvirtError(Exception):
        """
        libvirtError mock
        """

        def __init__(self, msg):
            super().__init__(msg)
            self.msg = msg

        def get_error_message(self):
            return self.msg


@pytest.fixture(autouse=True)
def setup_loader():
    setup_loader_modules = {virt: {"libvirt": LibvirtMock()}}
    with pytest.helpers.loader_mock(setup_loader_modules) as loader_mock:
        yield loader_mock


@pytest.fixture(params=[True, False], ids=["test", "notest"])
def test(request):
    """
    Run the test with both True and False test values
    """
    return request.param



# Author: Bo Maryniuk <bo@suse.de>

import pytest
import salt.modules.ansiblegate as ansiblegate
import salt.utils.json
from tests.support.mock import ANY, MagicMock, patch

pytestmark = [
    pytest.mark.skip_on_windows(reason="Not supported on Windows"),
]


@pytest.fixture
def configure_loader_modules():
    return {ansiblegate: {}}


def test_ansible_module_help():
    """
    Test help extraction from the module
    :return:
    """
    extension = {
        "foo": {
            "doc": {"description": "The description of foo"},
            "examples": "These are the examples",
            "return": {"a": "A return"},
        }
    }

    with patch("subprocess.run") as proc_run_mock:
        proc_run_mock.return_value.stdout = salt.utils.json.dumps(extension)
        ret = ansiblegate.help("foo")
        assert ret["description"] == extension["foo"]["doc"]["description"]


def test_virtual_function(subtests):
    """
    Test Ansible module __virtual__ when ansible is not installed on the minion.
    :return:
    """

    with subtests.test("missing ansible binary"):
        with patch("salt.utils.path.which", side_effect=[None]):
            assert ansiblegate.__virtual__() == (
                False,
                "The 'ansible' binary was not found.",
            )

    with subtests.test("missing ansible-doc binary"):
        with patch(
            "salt.utils.path.which",
            side_effect=["/path/to/ansible", None],
        ):
            assert ansiblegate.__virtual__() == (
                False,
                "The 'ansible-doc' binary was not found.",
            )

    with subtests.test("missing ansible-playbook binary"):
        with patch(
            "salt.utils.path.which",
            side_effect=["/path/to/ansible", "/path/to/ansible-doc", None],
        ):
            assert ansiblegate.__virtual__() == (
                False,
                "The 'ansible-playbook' binary was not found.",
            )

    with subtests.test("Failing to load the ansible modules listing"):
        with patch(
            "salt.utils.path.which",
            side_effect=[
                "/path/to/ansible",
                "/path/to/ansible-doc",
                "/path/to/ansible-playbook",
            ],
        ):
            with patch("subprocess.run") as proc_run_mock:
                proc_run_mock.return_value.retcode = 1
                proc_run_mock.return_value.stderr = "bar"
                proc_run_mock.return_value.stdout = "{}"
                assert ansiblegate.__virtual__() == (
                    False,
                    "Failed to get the listing of ansible modules:\nbar",
                )


def test_ansible_module_call():
    """
    Test Ansible module call from ansible gate module
    :return:
    """

    with patch("subprocess.run") as proc_run_mock:
        proc_run_mock.return_value.stdout = (
            'localhost | SUCCESS => {\n    "completed": true    \n}'
        )

        ret = ansiblegate.call("one.two.three", "arg_1", kwarg1="foobar")
        proc_run_mock.assert_any_call(
            [
                ANY,
                "localhost",
                "--limit",
                "127.0.0.1",
                "-m",
                "one.two.three",
                "-a",
                '"arg_1" kwarg1="foobar"',
                "-i",
                ANY,
            ],
            check=True,
            shell=False,
            stderr=-1,
            stdout=-1,
            timeout=1200,
            universal_newlines=True,
        )
        assert ret == {"completed": True}


def test_ansible_playbooks_return_retcode():
    """
    Test ansible.playbooks execution module function include retcode in the return.
    :return:
    """
    ref_out = {"retcode": 0, "stdout": '{"foo": "bar"}'}
    cmd_run_all = MagicMock(return_value=ref_out)
    with patch.dict(ansiblegate.__salt__, {"cmd.run_all": cmd_run_all}), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = ansiblegate.playbooks("fake-playbook.yml")
        assert "retcode" in ret



import pytest

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_requisites_mixed_require_prereq_use_1(state, state_tree):
    """
    Call sls file containing several requisites.
    """
    expected_simple_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo C" run',
            "result": True,
            "changes": True,
        },
    }
    sls_contents = """
    # Simple mix between prereq and require
    # C (1) <--+ <------+
    #          |        |
    # B (2) -p-+ <-+    |
    #              |    |
    # A (3) --r----+ -p-+

    A:
      cmd.run:
        - name: echo A
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C
    B:
      cmd.run:
        - name: echo B
        - require_in:
          - cmd: A

    # infinite recursion.....?
    C:
      cmd.run:
        - name: echo C
        # will test B and be applied only if B changes,
        # and then will run before B
        - prereq:
            - cmd: B
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_simple_result


@pytest.mark.skip("Undetected infinite loops prevents this test from running...")
def test_requisites_mixed_require_prereq_use_2(state, state_tree):
    sls_contents = """
    # Complex require/require_in/prereq/preqreq_in graph
    #
    #
    # D (1) <--------r-----+
    #                      |
    # C (2) <--+ <-----p-------+
    #          |           |   |
    # B (3) -p-+ <-+ <-+ --+   |
    #              |   |       |
    # E (4) ---r---|---+ <-+   |
    #              |       |   |
    # A (5) --r----+ ---r--+ --+
    #

    A:
      cmd.run:
        - name: echo A fifth
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C
    B:
      cmd.run:
        - name: echo B third
        - require_in:
          - cmd: A

    # infinite recursion.....
    C:
      cmd.run:
        - name: echo C second
        # will test B and be applied only if B changes,
        # and then will run before B
        - prereq:
            - cmd: B

    D:
      cmd.run:
        - name: echo D first
        # issue #8773
        - require_in:
            cmd: B

    E:
      cmd.run:
        - name: echo E fourth
        - require:
          - cmd: B
        - require_in:
          - cmd: A
    """
    expected_result = {
        "cmd_|-A_|-echo A fifth_|-run": {
            "__run_num__": 4,
            "comment": 'Command "echo A fifth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo B third" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo C second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-D_|-echo D first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo D first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-E_|-echo E fourth_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo E fourth" run',
            "result": True,
            "changes": True,
        },
    }
    # undetected infinite loops prevents this test from running...
    # TODO: this is actually failing badly
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


@pytest.mark.skip("Undetected infinite loops prevents this test from running...")
def test_requisites_mixed_require_prereq_use_3(state, state_tree):
    # test Traceback recursion prereq+require #8785
    sls_contents = """
    # issue #8785
    # B <--+ ----r-+
    #      |       |
    # A -p-+ <-----+-- ERROR: cannot respect both require and prereq

    A:
      cmd.run:
        - name: echo A
        - require_in:
          - cmd: B

    # infinite recursion.....?
    B:
      cmd.run:
        - name: echo B
        # will test A and be applied only if A changes,
        # and then will run before A
        - prereq:
            - cmd: A
    """
    expected_result = ['A recursive requisite was found, SLS "requisite" ID "B" ID "A"']
    # TODO: this is actually failing badly
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert isinstance(ret, list)  # Error
        assert ret == expected_result


@pytest.mark.skip("Undetected infinite loops prevents this test from running...")
def test_requisites_mixed_require_prereq_use_4(state, state_tree):
    # test Infinite recursion prereq+require #8785 v2
    sls_contents = """
    # issue #8785 RuntimeError: maximum recursion depth exceeded
    # C <--+ <------+ -r-+
    #      |        |    |
    # B -p-+ <-+    | <--+-- ERROR: cannot respect both require and prereq
    #          |    |
    # A --r----+ -p-+

    A:
      cmd.run:
        - name: echo A
        # is running in test mode before C
        # C gets executed first if this states modify something
        - prereq_in:
          - cmd: C
    B:
      cmd.run:
        - name: echo B
        - require_in:
          - cmd: A
          # this should raise the error
          - cmd: C

    # infinite recursion.....?
    C:
      cmd.run:
        - name: echo C
        # will test B and be applied only if B changes,
        # and then will run before B
        - prereq:
            - cmd: B
    """
    expected_result = ['A recursive requisite was found, SLS "requisite" ID "B" ID "A"']
    # TODO: this is actually failing badly
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert isinstance(ret, list)  # Error
        assert ret == expected_result


@pytest.mark.skip("Undetected infinite loops prevents this test from running...")
def test_requisites_mixed_require_prereq_use_5(state, state_tree):
    # test Infinite recursion prereq+require #8785 v3
    sls_contents = """
    # issue #8785
    #
    # Here it's more complex. Order SHOULD be ok.
    # When B changes something the require is verified.
    # What should happen if B does not chane anything?
    # It should also run because of the require.
    # Currently we have:
    # RuntimeError: maximum recursion depth exceeded

    # B (1) <---+ <--+
    #           |    |
    # A (2) -r--+ -p-+

    A:
      cmd.run:
        - name: echo A
        # is running in test mode before B
        # B gets executed first if this states modify something
        # key of bug
        - prereq_in:
          - cmd: B
    B:
      cmd.run:
        - name: echo B
        # B should run before A
        - require_in:
          - cmd: A
    """
    expected_result = ['A recursive requisite was found, SLS "requisite" ID "B" ID "A"']
    # TODO: this is actually failing badly, and expected result is maybe not a recursion
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert isinstance(ret, list)  # Error
        assert ret == expected_result


def test_issue_46762_prereqs_on_a_state_with_unfulfilled_requirements(
    state, state_tree
):
    """
    This tests the case where state C requires state A, which fails.
    State C is a pre-required state for State B.
    Since state A fails, state C will not run because the requisite failed,
    therefore state B will not run because state C failed to run.

    See https://github.com/saltstack/salt/issues/46762 for
    more information.
    """
    sls_contents = """
    a: test.fail_without_changes

    b:
      test.nop:
        - prereq:
          - c

    c:
      test.nop:
      - require:
        - a
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")

    state_id = "test_|-a_|-a_|-fail_without_changes"
    assert state_id in ret
    assert ret[state_id].result is False
    assert ret[state_id].comment == "Failure!"

    state_id = "test_|-b_|-b_|-nop"
    assert state_id in ret
    assert ret[state_id].result is False
    assert ret[state_id].comment == "One or more requisite failed: requisite.c"

    state_id = "test_|-c_|-c_|-nop"
    assert state_id in ret
    assert ret[state_id].result is False
    assert ret[state_id].comment == "One or more requisite failed: requisite.a"


@pytest.mark.skip_on_darwin(reason="Test is broken on macosx")
@pytest.mark.slow_test
def test_issue_30161_unless_and_onlyif_together(state, state_tree, tmp_path):
    """
    test cmd.run using multiple unless options where the first cmd in the
    list will pass, but the second will fail. This tests the fix for issue
    #35384. (The fix is in PR #35545.)
    """
    test_txt_path = tmp_path / "test.txt"
    sls_contents = """
    unless_false_onlyif_true:
      file.managed:
        - name: {test_txt_path}
        - unless: {test_false}
        - onlyif: {test_true}

    unless_true_onlyif_false:
      file.managed:
        - name: {test_txt_path}
        - unless: {test_true}
        - onlyif: {test_false}

    unless_true_onlyif_true:
      file.managed:
        - name: {test_txt_path}
        - unless: {test_true}
        - onlyif: {test_true}

    unless_false_onlyif_false:
      file.managed:
        - name: {test_txt_path}
        - contents: test
        - unless: {test_false}
        - onlyif: {test_false}
    """.format(
        test_true=pytest.helpers.shell_test_true(),
        test_false=pytest.helpers.shell_test_false(),
        test_txt_path=test_txt_path,
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        for state_entry in ret:
            assert state_entry.result is True
        # We must assert against the comment here to make sure the comment reads that the
        # command "echo "hello"" was run. This ensures that we made it to the last unless
        # command in the state. If the comment reads "unless condition is true", or similar,
        # then the unless state run bailed out after the first unless command succeeded,
        # which is the bug we're regression testing for.
    _expected = {
        "file_|-unless_false_onlyif_false_|-{}_|-managed".format(test_txt_path): {
            "comment": "onlyif condition is false\nunless condition is false",
            "name": "{}".format(test_txt_path),
            "skip_watch": True,
            "changes": {},
            "result": True,
        },
        "file_|-unless_false_onlyif_true_|-{}_|-managed".format(test_txt_path): {
            "comment": "Empty file",
            "name": str(test_txt_path),
            "start_time": "18:10:20.341753",
            "result": True,
            "changes": {"new": "file {} created".format(test_txt_path)},
        },
        "file_|-unless_true_onlyif_false_|-{}_|-managed".format(test_txt_path): {
            "comment": "onlyif condition is false\nunless condition is true",
            "name": str(test_txt_path),
            "start_time": "18:10:22.936446",
            "skip_watch": True,
            "changes": {},
            "result": True,
        },
        "file_|-unless_true_onlyif_true_|-{}_|-managed".format(test_txt_path): {
            "comment": "onlyif condition is true\nunless condition is true",
            "name": str(test_txt_path),
            "skip_watch": True,
            "changes": {},
            "result": True,
        },
    }
    for slsid in _expected:
        assert ret[slsid].comment == _expected[slsid]["comment"]



"""
Tests for the MySQL states
"""

import logging

import pytest
import salt.utils.path
from salt.modules import mysql as mysqlmod
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf

log = logging.getLogger(__name__)

NO_MYSQL = False
try:
    import MySQLdb  # pylint: disable=import-error,unused-import
except ImportError:
    NO_MYSQL = True

if not salt.utils.path.which("mysqladmin"):
    NO_MYSQL = True


@skipIf(
    NO_MYSQL,
    "Please install MySQL bindings and a MySQL Server before running"
    "MySQL integration tests.",
)
class MysqlGrantsStateTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the mysql_grants states
    """

    user = "root"
    password = "poney"
    # yep, theses are valid MySQL db names
    # very special chars are _ % and .
    testdb1 = "tes.t'\"saltdb"
    testdb2 = "t_st `(:=salt%b)"
    testdb3 = "test `(:=salteeb)"
    table1 = "foo"
    table2 = "foo `'%_bar"
    users = {
        "user1": {"name": "foo", "pwd": "bar"},
        "user2": {"name": 'user ";--,?:&/\\', "pwd": '";--(),?:@=&/\\'},
        # this is : passwd 標標
        "user3": {"name": "user( @ )=foobar", "pwd": "\xe6\xa8\x99\xe6\xa8\x99"},
        # this is : user/password containing 標標
        "user4": {"name": "user \xe6\xa8\x99", "pwd": "\xe6\xa8\x99\xe6\xa8\x99"},
    }

    @pytest.mark.destructive_test
    def setUp(self):
        """
        Test presence of MySQL server, enforce a root password
        """
        super().setUp()
        NO_MYSQL_SERVER = True
        # now ensure we know the mysql root password
        # one of theses two at least should work
        ret1 = self.run_state(
            "cmd.run",
            name='mysqladmin --host="localhost" -u '
            + self.user
            + ' flush-privileges password "'
            + self.password
            + '"',
        )
        ret2 = self.run_state(
            "cmd.run",
            name='mysqladmin --host="localhost" -u '
            + self.user
            + ' --password="'
            + self.password
            + '" flush-privileges password "'
            + self.password
            + '"',
        )
        key, value = ret2.popitem()
        if value["result"]:
            NO_MYSQL_SERVER = False
        else:
            self.skipTest("No MySQL Server running, or no root access on it.")
        # Create some users and a test db
        for user, userdef in self.users.items():
            self._userCreation(uname=userdef["name"], password=userdef["pwd"])
        self.run_state(
            "mysql_database.present",
            name=self.testdb1,
            character_set="utf8",
            collate="utf8_general_ci",
            connection_user=self.user,
            connection_pass=self.password,
        )
        self.run_state(
            "mysql_database.present",
            name=self.testdb2,
            character_set="utf8",
            collate="utf8_general_ci",
            connection_user=self.user,
            connection_pass=self.password,
        )
        create_query = (
            "CREATE TABLE {tblname} ("
            " id INT NOT NULL AUTO_INCREMENT PRIMARY KEY,"
            " data VARCHAR(100)) ENGINE={engine};".format(
                tblname=mysqlmod.quote_identifier(self.table1),
                engine="MYISAM",
            )
        )
        log.info("Adding table '%s'", self.table1)
        self.run_function(
            "mysql.query",
            database=self.testdb2,
            query=create_query,
            connection_user=self.user,
            connection_pass=self.password,
        )
        create_query = (
            "CREATE TABLE {tblname} ("
            " id INT NOT NULL AUTO_INCREMENT PRIMARY KEY,"
            " data VARCHAR(100)) ENGINE={engine};".format(
                tblname=mysqlmod.quote_identifier(self.table2),
                engine="MYISAM",
            )
        )
        log.info("Adding table '%s'", self.table2)
        self.run_function(
            "mysql.query",
            database=self.testdb2,
            query=create_query,
            connection_user=self.user,
            connection_pass=self.password,
        )

    @pytest.mark.destructive_test
    def tearDown(self):
        """
        Removes created users and db
        """
        for user, userdef in self.users.items():
            self._userRemoval(uname=userdef["name"], password=userdef["pwd"])
        self.run_state(
            "mysql_database.absent",
            name=self.testdb1,
            connection_user=self.user,
            connection_pass=self.password,
        )
        self.run_function(
            "mysql_database.absent",
            name=self.testdb2,
            connection_user=self.user,
            connection_pass=self.password,
        )

    def _userCreation(self, uname, password=None):
        """
        Create a test user
        """
        self.run_state(
            "mysql_user.present",
            name=uname,
            host="localhost",
            password=password,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
            saltenv={"LC_ALL": "en_US.utf8"},
        )

    def _userRemoval(self, uname, password=None):
        """
        Removes a test user
        """
        self.run_state(
            "mysql_user.absent",
            name=uname,
            host="localhost",
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
            saltenv={"LC_ALL": "en_US.utf8"},
        )

    @pytest.mark.destructive_test
    def test_grant_present_absent(self):
        """
        mysql_database.present
        """
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 1",
            grant="SELECT, INSERT",
            database=self.testdb1 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 2",
            grant="SELECT, ALTER,CREATE TEMPORARY tables, execute",
            database=self.testdb1 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 3",
            grant="SELECT, INSERT",
            database=self.testdb2 + "." + self.table2,
            user=self.users["user2"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 4",
            grant="SELECT, INSERT",
            database=self.testdb2 + "." + self.table2,
            user=self.users["user2"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 5",
            grant="SELECT, UPDATE",
            database=self.testdb2 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=False,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.absent",
            name="grant test 6",
            grant="SELECT,update",
            database=self.testdb2 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=False,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest
import salt.states.timezone as timezone
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {timezone: {}}


def test_system():
    """
    Test to set the timezone for the system.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(side_effect=[CommandExecutionError, True, True, True])
    mock1 = MagicMock(side_effect=["local", "localtime", "localtime"])
    mock2 = MagicMock(return_value=False)
    with patch.dict(
        timezone.__salt__,
        {
            "timezone.zone_compare": mock,
            "timezone.get_hwclock": mock1,
            "timezone.set_hwclock": mock2,
        },
    ):
        ret.update(
            {
                "comment": (
                    "Unable to compare desired timezone 'salt' to system timezone: "
                ),
                "result": False,
            }
        )
        assert timezone.system("salt") == ret

        ret.update(
            {
                "comment": "Timezone salt already set, UTC already set to salt",
                "result": True,
            }
        )
        assert timezone.system("salt") == ret

        with patch.dict(timezone.__opts__, {"test": True}):
            ret.update({"comment": "UTC needs to be set to True", "result": None})
            assert timezone.system("salt") == ret

        with patch.dict(timezone.__opts__, {"test": False}):
            ret.update({"comment": "Failed to set UTC to True", "result": False})
            assert timezone.system("salt") == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.rdp as rdp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rdp: {}}


def test_enabled():
    """
    Test to enable the RDP service and make sure access
    to the RDP port is allowed in the firewall configuration.
    """
    name = "my_service"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock_t = MagicMock(side_effect=[False, False, True])
    mock_f = MagicMock(return_value=False)
    with patch.dict(rdp.__salt__, {"rdp.status": mock_t, "rdp.enable": mock_f}):
        with patch.dict(rdp.__opts__, {"test": True}):
            comt = "RDP will be enabled"
            ret.update({"comment": comt, "result": None})
            assert rdp.enabled(name) == ret

        with patch.dict(rdp.__opts__, {"test": False}):
            ret.update(
                {"comment": "", "result": False, "changes": {"RDP was enabled": True}}
            )
            assert rdp.enabled(name) == ret

            comt = "RDP is enabled"
            ret.update({"comment": comt, "result": True, "changes": {}})
            assert rdp.enabled(name) == ret


def test_disabled():
    """
    Test to disable the RDP service.
    """
    name = "my_service"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(side_effect=[True, True, False])
    mock_t = MagicMock(return_value=True)
    with patch.dict(rdp.__salt__, {"rdp.status": mock, "rdp.disable": mock_t}):
        with patch.dict(rdp.__opts__, {"test": True}):
            comt = "RDP will be disabled"
            ret.update({"comment": comt, "result": None})
            assert rdp.disabled(name) == ret

        with patch.dict(rdp.__opts__, {"test": False}):
            ret.update(
                {"comment": "", "result": True, "changes": {"RDP was disabled": True}}
            )
            assert rdp.disabled(name) == ret

            comt = "RDP is disabled"
            ret.update({"comment": comt, "result": True, "changes": {}})
            assert rdp.disabled(name) == ret



"""
Junos redundant routing engine beacon.

.. note::

   This beacon only works on the Juniper native minion.

Copies salt-minion keys to the backup RE when present

Configure with

.. code-block:: yaml

    beacon:
      beacons:
        junos_rre_keys:
          - interval: 43200

`interval` above is in seconds, 43200 is recommended (every 12 hours)
"""

__virtualname__ = "junos_rre_keys"


def beacon(config):
    ret = []

    engine_status = __salt__["junos.routing_engine"]()

    if not engine_status["success"]:
        return []

    for e in engine_status["backup"]:
        result = __salt__["junos.dir_copy"]("/var/local/salt/etc", e)
        ret.append({"result": result, "success": True})

    return ret



"""
    Unit tests for salt.utils.yamldumper
"""

import salt.utils.yamldumper
from tests.support.unit import TestCase


class YamlDumperTestCase(TestCase):
    """
    TestCase for salt.utils.yamldumper module
    """

    def test_yaml_dump(self):
        """
        Test yaml.dump a dict
        """
        data = {"foo": "bar"}
        exp_yaml = "{foo: bar}\n"

        assert salt.utils.yamldumper.dump(data) == exp_yaml

        assert salt.utils.yamldumper.dump(
            data, default_flow_style=False
        ) == exp_yaml.replace("{", "").replace("}", "")

    def test_yaml_safe_dump(self):
        """
        Test yaml.safe_dump a dict
        """
        data = {"foo": "bar"}
        assert salt.utils.yamldumper.safe_dump(data) == "{foo: bar}\n"

        assert (
            salt.utils.yamldumper.safe_dump(data, default_flow_style=False)
            == "foo: bar\n"
        )



import logging

import pytest
from tests.support.mock import MagicMock, patch

try:
    import dateutil.parser as dateutil_parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        HAS_DATEUTIL_PARSER is False,
        reason="The 'dateutil.parser' library is not available",
    ),
    pytest.mark.windows_whitelisted,
]


def test_maxrunning_minion(schedule):
    """
    verify that scheduled job runs
    """
    schedule.opts["__role"] = "minion"

    job = {
        "schedule": {
            "maxrunning_minion": {
                "function": "test.ping",
                "seconds": 10,
                "maxrunning": 1,
            }
        }
    }

    job_data = {
        "function": "test.ping",
        "run": True,
        "name": "maxrunning_minion",
        "seconds": 10,
        "_seconds": 10,
        "jid_include": True,
        "maxrunning": 1,
    }

    # Add the job to the scheduler
    schedule.opts.update(job)

    running_data = [
        {
            "fun_args": [],
            "jid": "20181018165923360935",
            "schedule": "maxrunning_minion",
            "pid": 15338,
            "fun": "test.ping",
            "id": "host",
        }
    ]

    run_time = dateutil_parser.parse("11/29/2017 4:00pm")

    with patch("salt.utils.minion.running", MagicMock(return_value=running_data)):
        with patch("salt.utils.process.os_is_running", MagicMock(return_value=True)):
            ret = schedule._check_max_running(
                "test.ping", job_data, schedule.opts, now=run_time
            )
    assert "_skip_reason" in ret
    assert "maxrunning" == ret["_skip_reason"]
    assert not ret["run"]


def test_maxrunning_master(schedule):
    """
    verify that scheduled job runs
    """
    schedule.opts["__role"] = "master"

    job = {
        "schedule": {
            "maxrunning_master": {
                "function": "state.orch",
                "args": ["test.orch_test"],
                "minutes": 1,
                "maxrunning": 1,
            }
        }
    }

    job_data = {
        "function": "state.orch",
        "fun_args": ["test.orch_test"],
        "run": True,
        "name": "maxrunning_master",
        "minutes": 1,
        "jid_include": True,
        "maxrunning": 1,
    }

    # Add the job to the scheduler
    schedule.opts.update(job)

    running_data = [
        {
            "fun_args": ["test.orch_test"],
            "jid": "20181018165923360935",
            "schedule": "maxrunning_master",
            "pid": 15338,
            "fun": "state.orch",
            "id": "host",
        }
    ]

    run_time = dateutil_parser.parse("11/29/2017 4:00pm")

    with patch(
        "salt.utils.master.get_running_jobs", MagicMock(return_value=running_data)
    ):
        with patch("salt.utils.process.os_is_running", MagicMock(return_value=True)):
            ret = schedule._check_max_running(
                "state.orch", job_data, schedule.opts, now=run_time
            )
    assert "_skip_reason" in ret
    assert "maxrunning" == ret["_skip_reason"]
    assert not ret["run"]



import shutil
import textwrap
import time

import attr
import pytest


@attr.s
class BlackoutPillar:
    pillar_state_tree = attr.ib(repr=False)
    salt_cli = attr.ib(repr=False)
    top_file = attr.ib(init=False)
    minion_1_id = attr.ib(repr=False)
    minion_1_pillar = attr.ib(init=False)
    minion_2_id = attr.ib(repr=False)
    minion_2_pillar = attr.ib(init=False)
    in_blackout = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.top_file = self.pillar_state_tree / "top.sls"
        top_file_contents = textwrap.dedent(
            """\
        base:
          {}:
            - minion-1-pillar
          {}:
            - minion-2-pillar
        """.format(
                self.minion_1_id, self.minion_2_id
            )
        )
        self.top_file.write_text(top_file_contents)
        self.minion_1_pillar = self.pillar_state_tree / "minion-1-pillar.sls"
        self.minion_2_pillar = self.pillar_state_tree / "minion-2-pillar.sls"
        self.minion_1_pillar.write_text("minion_blackout: false")
        self.minion_2_pillar.write_text("minion_blackout: false")
        self.refresh_pillar()

    def enter_blackout(self, pillar_contents=None):
        if pillar_contents is None:
            pillar_contents = "minion_blackout: false"
        if pillar_contents.startswith("\n"):
            pillar_contents = pillar_contents[1:]
        pillar_contents = textwrap.dedent(pillar_contents)
        self.minion_1_pillar.write_text(pillar_contents)
        self.refresh_pillar(exiting_blackout=False)
        self.in_blackout = True
        return self.__enter__()

    def exit_blackout(self):
        if self.in_blackout:
            self.minion_1_pillar.write_text("minion_blackout: false")
            self.refresh_pillar(exiting_blackout=True)
            self.in_blackout = False

    def refresh_pillar(self, timeout=60, sleep=0.5, exiting_blackout=None):
        ret = self.salt_cli.run("saltutil.refresh_pillar", wait=True, minion_tgt="*")
        assert ret.exitcode == 0
        assert self.minion_1_id in ret.json
        assert self.minion_2_id in ret.json
        stop_at = time.time() + timeout
        while True:
            if time.time() > stop_at:
                if exiting_blackout is True:
                    pytest.fail(
                        "Minion did not exit blackout mode after {} seconds".format(
                            timeout
                        )
                    )
                elif exiting_blackout is False:
                    pytest.fail(
                        "Minion did not enter blackout mode after {} seconds".format(
                            timeout
                        )
                    )
                else:
                    pytest.fail(
                        "Minion did not refresh pillar after {} seconds".format(timeout)
                    )

            time.sleep(sleep)

            ret = self.salt_cli.run("pillar.get", "minion_blackout", minion_tgt="*")
            if not ret.json:
                # Something is wrong here. Try again
                continue
            assert self.minion_1_id in ret.json
            assert self.minion_2_id in ret.json
            if ret.json[self.minion_1_id] == "" or ret.json[self.minion_2_id] == "":
                # Pillar not found
                continue

            # Minion 2 must NEVER enter blackout
            assert ret.json[self.minion_2_id] is False

            if exiting_blackout is True and ret.json[self.minion_1_id] is not False:
                continue
            elif (
                exiting_blackout is False
                and "Minion in blackout mode" not in ret.json[self.minion_1_id]
            ):
                continue
            # We got the pillar we're after, break out of the loop
            break

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.exit_blackout()


@pytest.fixture(scope="package")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="package")
def salt_master(salt_factories, pillar_state_tree):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
    }
    factory = salt_factories.salt_master_daemon(
        "blackout-master",
        defaults=config_defaults,
        extra_cli_arguments_after_first_start_failure=["--log-level=debug"],
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="package")
def salt_minion_1(salt_master):
    factory = salt_master.salt_minion_daemon(
        "blackout-minion-1", defaults={"open_mode": True}
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="package")
def salt_minion_2(salt_master):
    factory = salt_master.salt_minion_daemon(
        "blackout-minion-2", defaults={"open_mode": True}
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="package")
def salt_cli(salt_master):
    return salt_master.salt_cli()


@pytest.fixture(scope="package")
def blackout_pillar(salt_cli, pillar_state_tree, salt_minion_1, salt_minion_2):
    return BlackoutPillar(
        pillar_state_tree=pillar_state_tree,
        salt_cli=salt_cli,
        minion_1_id=salt_minion_1.id,
        minion_2_id=salt_minion_2.id,
    )


@pytest.fixture
def blackout(blackout_pillar):
    try:
        yield blackout_pillar
    finally:
        blackout_pillar.exit_blackout()



"""
Unit tests for salt/modules/salt_version.py
"""


import salt.modules.salt_version as salt_version
import salt.version
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SaltVersionTestCase(TestCase):
    """
    Test cases for salt.modules.salt_version
    """

    def test_mocked_objects(self):
        """
        Test that the mocked objects actually have what we expect.

        For example, earlier tests incorrectly mocked the
        salt.version.SaltStackVersion.LNAMES dict using upper-case indexes
        """
        assert isinstance(salt.version.SaltStackVersion.LNAMES, dict)
        sv = salt.version.SaltStackVersion(*salt.version.__version_info__)
        for k, v in salt.version.SaltStackVersion.LNAMES.items():
            assert k == k.lower()
            assert isinstance(v, tuple)
            if sv.new_version(major=v[0]):
                assert len(v) == 1
            else:
                assert len(v) == 2

        sv = sv.__str__()
        assert isinstance(sv, str)

        with patch("salt.version.SaltStackVersion.LNAMES", {"neon": (2019, 8)}):
            sv = salt.version.SaltStackVersion.from_name("Neon")
            self.assertEqual(sv.string, "2019.8.0")

    # get_release_number tests: 3

    def test_get_release_number_no_codename(self):
        """
        Test that None is returned when the codename isn't found.
        """
        assert salt_version.get_release_number("foo") is None

    @patch("salt.version.SaltStackVersion.LNAMES", {"foo": (12345, 0)})
    def test_get_release_number_unassigned(self):
        """
        Test that a string is returned when a version is found, but unassigned.
        """
        mock_str = "No version assigned."
        assert salt_version.get_release_number("foo") == mock_str

    def test_get_release_number_success(self):
        """
        Test that a version is returned for a released codename
        """
        assert salt_version.get_release_number("Oxygen") == "2018.3"

    def test_get_release_number_success_new_version(self):
        """
        Test that a version is returned for new versioning (3000)
        """
        assert salt_version.get_release_number("Neon") == "3000"

    # equal tests: 3

    @patch("salt.version.SaltStackVersion.LNAMES", {"foo": (1900, 5)})
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="1900.5.0"))
    def test_equal_success(self):
        """
        Test that the current version is equal to the codename
        """
        assert salt_version.equal("foo") is True

    @patch("salt.version.SaltStackVersion.LNAMES", {"foo": (3000,)})
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="3000.1"))
    def test_equal_success_new_version(self):
        """
        Test that the current version is equal to the codename
        while using the new versioning
        """
        assert salt_version.equal("foo") is True

    @patch(
        "salt.version.SaltStackVersion.LNAMES",
        {"oxygen": (2018, 3), "nitrogen": (2017, 7)},
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_equal_older_codename(self):
        """
        Test that when an older codename is passed in, the function returns False.
        """
        assert salt_version.equal("Nitrogen") is False

    @patch(
        "salt.version.SaltStackVersion.LNAMES", {"neon": (3000), "nitrogen": (2017, 7)}
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_equal_older_codename_new_version(self):
        """
        Test that when an older codename is passed in, the function returns False.
        while also testing with the new versioning.
        """
        assert salt_version.equal("Nitrogen") is False

    @patch(
        "salt.version.SaltStackVersion.LNAMES",
        {"fluorine": (salt.version.MAX_SIZE - 100, 0)},
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_equal_newer_codename(self):
        """
        Test that when a newer codename is passed in, the function returns False
        """
        assert salt_version.equal("Fluorine") is False

    # greater_than tests: 4

    @patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2017.7")
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_greater_than_success(self):
        """
        Test that the current version is newer than the codename
        """
        assert salt_version.greater_than("Nitrogen") is True

    @patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2017.7")
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="3000"))
    def test_greater_than_success_new_version(self):
        """
        Test that the current version is newer than the codename
        """
        assert salt_version.greater_than("Nitrogen") is True

    @patch("salt.version.SaltStackVersion.LNAMES", {"oxygen": (2018, 3)})
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_greater_than_with_equal_codename(self):
        """
        Test that when an equal codename is passed in, the function returns False.
        """
        assert salt_version.greater_than("Oxygen") is False

    @patch(
        "salt.version.SaltStackVersion.LNAMES",
        {"fluorine": (2019, 2), "oxygen": (2018, 3)},
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_greater_than_with_newer_codename(self):
        """
        Test that when a newer codename is passed in, the function returns False.
        """
        assert salt_version.greater_than("Fluorine") is False

    @patch(
        "salt.modules.salt_version.get_release_number",
        MagicMock(return_value="No version assigned."),
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_greater_than_unassigned(self):
        """
        Test that the unassigned codename is greater than the current version
        """
        assert salt_version.greater_than("Fluorine") is False

    # less_than tests: 4

    @patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2019.2")
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_less_than_success(self):
        """
        Test that when a newer codename is passed in, the function returns True.
        """
        assert salt_version.less_than("Fluorine") is True

    @patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="3000")
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_less_than_success_new_version(self):
        """
        Test that when a newer codename is passed in, the function returns True
        using new version
        """
        assert salt_version.less_than("Fluorine") is True

    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    @patch("salt.version.SaltStackVersion.LNAMES", {"oxygen": (2018, 3)})
    def test_less_than_with_equal_codename(self):
        """
        Test that when an equal codename is passed in, the function returns False.
        """
        assert salt_version.less_than("Oxygen") is False

    @patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2017.7")
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_less_than_with_older_codename(self):
        """
        Test that the current version is less than the codename.
        """
        assert salt_version.less_than("Nitrogen") is False

    @patch(
        "salt.modules.salt_version.get_release_number",
        MagicMock(return_value="No version assigned."),
    )
    @patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2"))
    def test_less_than_with_unassigned_codename(self):
        """
        Test that when an unassigned codename greater than the current version.
        """
        assert salt_version.less_than("Fluorine") is True

    # _check_release_cmp tests: 2

    def test_check_release_cmp_no_codename(self):
        """
        Test that None is returned when the codename isn't found.
        """
        assert salt_version._check_release_cmp("foo") is None

    def test_check_release_cmp_success(self):
        """
        Test that an int is returned from the version compare
        """
        assert isinstance(salt_version._check_release_cmp("Oxygen"), int)



"""
Writes matches to disk to verify activity, helpful when testing

Normally this is used by giving the name of the file (without a path) that the
data will be saved to. If for instance you use ``foo`` as the name:

.. code-block:: yaml

    foo:
      file.save

Then the file will be saved to:

.. code-block:: bash

    <salt cachedir>/thorium/saves/foo

You may also provide an absolute path for the file to be saved to:

.. code-block:: yaml

    /tmp/foo.save:
        file.save

Files will be saved in JSON format. However, JSON does not support ``set()``s.
If you are saving a register entry that contains a ``set()``, then it will fail
to save to JSON format. However, you may pass data through a filter which makes
it JSON compliant:

.. code-block:: yaml

    foo:
      file.save:
        filter: True

Be warned that if you do this, then the file will be saved, but not in a format
that can be re-imported into Python.
"""


import os

import salt.utils.data
import salt.utils.files
import salt.utils.json


def save(name, filter=False):
    """
    Save the register to <salt cachedir>/thorium/saves/<name>, or to an
    absolute path.

    If an absolute path is specified, then the directory will be created
    non-recursively if it doesn't exist.

    USAGE:

    .. code-block:: yaml

        foo:
          file.save

        /tmp/foo:
          file.save
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if name.startswith("/"):
        tgt_dir = os.path.dirname(name)
        fn_ = name
    else:
        tgt_dir = os.path.join(__opts__["cachedir"], "thorium", "saves")
        fn_ = os.path.join(tgt_dir, name)
    if not os.path.isdir(tgt_dir):
        os.makedirs(tgt_dir)
    with salt.utils.files.fopen(fn_, "w+") as fp_:
        if filter is True:
            salt.utils.json.dump(salt.utils.data.simple_types_filter(__reg__), fp_)
        else:
            salt.utils.json.dump(__reg__, fp_)
    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.smf_service as smf
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SmfTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.smf
    """

    def setup_loader_modules(self):
        return {smf: {}}

    def test_get_running(self):
        """
        Test to return the running services
        """
        with patch.dict(
            smf.__salt__, {"cmd.run": MagicMock(return_value="A online\n")}
        ):
            self.assertEqual(smf.get_running(), ["A"])

    def test_get_stopped(self):
        """
        Test to return the stopped services
        """
        with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A\n")}):
            self.assertListEqual(smf.get_stopped(), ["A"])

    def test_available(self):
        """
        Test to returns ``True`` if the specified service is available,
        otherwise returns ``False``.
        """
        with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            with patch.object(smf, "get_all", return_value="A"):
                self.assertTrue(smf.available("A"))

    def test_missing(self):
        """
        The inverse of service.available.
        Returns ``True`` if the specified service is not available, otherwise
        returns ``False``.
        """
        with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            with patch.object(smf, "get_all", return_value="A"):
                self.assertFalse(smf.missing("A"))

    def test_get_all(self):
        """
        Test to return all installed services
        """
        with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A\n")}):
            self.assertListEqual(smf.get_all(), ["A"])

    def test_start(self):
        """
        Test to start the specified service
        """
        with patch.dict(
            smf.__salt__,
            {"cmd.retcode": MagicMock(side_effect=[False, 3, None, False, 4])},
        ):
            self.assertTrue(smf.start("name"))

            self.assertTrue(smf.start("name"))

            self.assertFalse(smf.start("name"))

    def test_stop(self):
        """
        Test to stop the specified service
        """
        with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
            self.assertTrue(smf.stop("name"))

    def test_restart(self):
        """
        Test to restart the named service
        """
        with patch.dict(
            smf.__salt__, {"cmd.retcode": MagicMock(side_effect=[False, True])}
        ):

            with patch.object(smf, "start", return_value="A"):
                self.assertEqual(smf.restart("name"), "A")

            self.assertFalse(smf.restart("name"))

    def test_reload_(self):
        """
        Test to reload the named service
        """
        with patch.dict(
            smf.__salt__, {"cmd.retcode": MagicMock(side_effect=[False, True])}
        ):

            with patch.object(smf, "start", return_value="A"):
                self.assertEqual(smf.reload_("name"), "A")

            self.assertFalse(smf.reload_("name"))

    def test_status(self):
        """
        Test to return the status for a service, returns a bool whether the
        service is running.
        """
        with patch.dict(
            smf.__salt__, {"cmd.run": MagicMock(side_effect=["online", "online1"])}
        ):
            self.assertTrue(smf.status("name"))

            self.assertFalse(smf.status("name"))

    def test_enable(self):
        """
        Test to enable the named service to start at boot
        """
        with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
            self.assertTrue(smf.enable("name"))

    def test_disable(self):
        """
        Test to disable the named service to start at boot
        """
        with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
            self.assertTrue(smf.disable("name"))

    def test_enabled(self):
        """
        Test to check to see if the named service is enabled to start on boot
        """
        with patch.dict(
            smf.__salt__,
            {
                "cmd.run": MagicMock(
                    side_effect=["fmri", "A B true", "fmri", "A B false"]
                )
            },
        ):
            self.assertTrue(smf.enabled("name"))

            self.assertFalse(smf.enabled("name"))

    def test_disabled(self):
        """
        Test to check to see if the named service is disabled to start on boot
        """
        with patch.object(smf, "enabled", return_value=False):
            self.assertTrue(smf.disabled("name"))

    def test_get_enabled(self):
        """
        Test to return the enabled services
        """
        with patch.object(smf, "_get_enabled_disabled", return_value=True):
            self.assertTrue(smf.get_enabled())

    def test_get_disabled(self):
        """
        Test to return the disabled services
        """
        with patch.object(smf, "_get_enabled_disabled", return_value=True):
            self.assertTrue(smf.get_disabled())



"""
Simple script to dump the contents of msgpack files to the terminal
"""
# pylint: disable=resource-leakage

import os
import pprint
import sys

import salt.utils.msgpack


def dump(path):
    """
    Read in a path and dump the contents to the screen
    """
    if not os.path.isfile(path):
        print("Not a file")
        return
    with open(path, "rb") as fp_:
        data = salt.utils.msgpack.loads(fp_.read())
        pprint.pprint(data)


if __name__ == "__main__":
    dump(sys.argv[1])



"""
Module to manage Linux kernel modules
"""

import logging
import os
import re

import salt.utils.files
import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only runs on Linux systems
    """
    return __grains__["kernel"] == "Linux"


def _new_mods(pre_mods, post_mods):
    """
    Return a list of the new modules, pass an lsmod dict before running
    modprobe and one after modprobe has run
    """
    pre = set()
    post = set()
    for mod in pre_mods:
        pre.add(mod["module"])
    for mod in post_mods:
        post.add(mod["module"])
    return post - pre


def _rm_mods(pre_mods, post_mods):
    """
    Return a list of the new modules, pass an lsmod dict before running
    modprobe and one after modprobe has run
    """
    pre = set()
    post = set()
    for mod in pre_mods:
        pre.add(mod["module"])
    for mod in post_mods:
        post.add(mod["module"])
    return pre - post


def _get_modules_conf():
    """
    Return location of modules config file.
    Default: /etc/modules
    """
    if "systemd" in __grains__:
        return "/etc/modules-load.d/salt_managed.conf"
    return "/etc/modules"


def _strip_module_name(mod):
    """
    Return module name and strip configuration. It is possible insert modules
    in this format:
        bonding mode=4 miimon=1000
    This method return only 'bonding'
    """
    if mod.strip() == "":
        return False
    return mod.split()[0]


def _set_persistent_module(mod):
    """
    Add module to configuration file to make it persistent. If module is
    commented uncomment it.
    """
    conf = _get_modules_conf()
    if not os.path.exists(conf):
        __salt__["file.touch"](conf)
    mod_name = _strip_module_name(mod)
    if not mod_name or mod_name in mod_list(True) or mod_name not in available():
        return set()
    escape_mod = re.escape(mod)
    # If module is commented only uncomment it
    if __salt__["file.search"](
        conf, "^#[\t ]*{}[\t ]*$".format(escape_mod), multiline=True
    ):
        __salt__["file.uncomment"](conf, escape_mod)
    else:
        __salt__["file.append"](conf, mod)
    return {mod_name}


def _remove_persistent_module(mod, comment):
    """
    Remove module from configuration file. If comment is true only comment line
    where module is.
    """
    conf = _get_modules_conf()
    mod_name = _strip_module_name(mod)
    if not mod_name or mod_name not in mod_list(True):
        return set()
    escape_mod = re.escape(mod)
    if comment:
        __salt__["file.comment"](conf, "^[\t ]*{}[\t ]?".format(escape_mod))
    else:
        __salt__["file.sed"](conf, "^[\t ]*{}[\t ]?".format(escape_mod), "")
    return {mod_name}


def available():
    """
    Return a list of all available kernel modules

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.available
    """
    ret = []

    mod_dir = os.path.join("/lib/modules/", os.uname()[2])

    built_in_file = os.path.join(mod_dir, "modules.builtin")
    if os.path.exists(built_in_file):
        with salt.utils.files.fopen(built_in_file, "r") as f:
            for line in f:
                # Strip .ko from the basename
                ret.append(os.path.basename(line)[:-4])

    for root, dirs, files in salt.utils.path.os_walk(mod_dir):
        for fn_ in files:
            if ".ko" in fn_:
                ret.append(fn_[: fn_.index(".ko")].replace("-", "_"))

    if "Arch" in __grains__["os_family"]:
        # Sadly this path is relative to kernel major version but ignores minor version
        mod_dir_arch = "/lib/modules/extramodules-" + os.uname()[2][0:3] + "-ARCH"
        for root, dirs, files in salt.utils.path.os_walk(mod_dir_arch):
            for fn_ in files:
                if ".ko" in fn_:
                    ret.append(fn_[: fn_.index(".ko")].replace("-", "_"))

    return sorted(list(ret))


def check_available(mod):
    """
    Check to see if the specified kernel module is available

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.check_available kvm
    """
    return mod in available()


def lsmod():
    """
    Return a dict containing information about currently loaded modules

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.lsmod
    """
    ret = []
    for line in __salt__["cmd.run"]("lsmod").splitlines():
        comps = line.split()
        if not len(comps) > 2:
            continue
        if comps[0] == "Module":
            continue
        mdat = {
            "size": comps[1],
            "module": comps[0],
            "depcount": comps[2],
        }
        if len(comps) > 3:
            mdat["deps"] = comps[3].split(",")
        else:
            mdat["deps"] = []
        ret.append(mdat)
    return ret


def mod_list(only_persist=False):
    """
    Return a list of the loaded module names

    only_persist
        Only return the list of loaded persistent modules

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.mod_list
    """
    mods = set()
    if only_persist:
        conf = _get_modules_conf()
        if os.path.exists(conf):
            try:
                with salt.utils.files.fopen(conf, "r") as modules_file:
                    for line in modules_file:
                        line = line.strip()
                        mod_name = _strip_module_name(line)
                        if not line.startswith("#") and mod_name:
                            mods.add(mod_name)
            except OSError:
                log.error("kmod module could not open modules file at %s", conf)
    else:
        for mod in lsmod():
            mods.add(mod["module"])
    return sorted(list(mods))


def load(mod, persist=False):
    """
    Load the specified kernel module

    mod
        Name of module to add

    persist
        Write module to /etc/modules to make it load on system reboot

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.load kvm
    """
    pre_mods = lsmod()
    res = __salt__["cmd.run_all"]("modprobe {}".format(mod), python_shell=False)
    if res["retcode"] == 0:
        post_mods = lsmod()
        mods = _new_mods(pre_mods, post_mods)
        persist_mods = set()
        if persist:
            persist_mods = _set_persistent_module(mod)
        return sorted(list(mods | persist_mods))
    else:
        return "Error loading module {}: {}".format(mod, res["stderr"])


def is_loaded(mod):
    """
    Check to see if the specified kernel module is loaded

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.is_loaded kvm
    """
    return mod in mod_list()


def remove(mod, persist=False, comment=True):
    """
    Remove the specified kernel module

    mod
        Name of module to remove

    persist
        Also remove module from /etc/modules

    comment
        If persist is set don't remove line from /etc/modules but only
        comment it

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.remove kvm
    """
    pre_mods = lsmod()
    res = __salt__["cmd.run_all"]("rmmod {}".format(mod), python_shell=False)
    if res["retcode"] == 0:
        post_mods = lsmod()
        mods = _rm_mods(pre_mods, post_mods)
        persist_mods = set()
        if persist:
            persist_mods = _remove_persistent_module(mod, comment)
        return sorted(list(mods | persist_mods))
    else:
        return "Error removing module {}: {}".format(mod, res["stderr"])



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.modules.win_ip as win_ip
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase

ETHERNET_CONFIG = (
    'Configuration for interface "Ethernet"\n'
    "DHCP enabled: Yes\n"
    "IP Address: 1.2.3.74\n"
    "Subnet Prefix: 1.2.3.0/24 (mask 255.255.255.0)\n"
    "Default Gateway: 1.2.3.1\n"
    "Gateway Metric: 0\n"
    "InterfaceMetric: 20\n"
    "DNS servers configured through DHCP: 1.2.3.4\n"
    "Register with which suffix: Primary only\n"
    "WINS servers configured through DHCP: None\n"
)

ETHERNET_ENABLE = (
    "Ethernet\nType: Dedicated\nAdministrative state: Enabled\nConnect state: Connected"
)


class WinShadowTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_ip
    """

    def setup_loader_modules(self):
        return {win_ip: {}}

    # 'raw_interface_configs' function tests: 1

    def test_raw_interface_configs(self):
        """
        Test if it return raw configs for all interfaces.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(win_ip.raw_interface_configs(), ETHERNET_CONFIG)

    # 'get_all_interfaces' function tests: 1

    def test_get_all_interfaces(self):
        """
        Test if it return configs for all interfaces.
        """
        ret = {
            "Ethernet": {
                "DHCP enabled": "Yes",
                "DNS servers configured through DHCP": ["1.2.3.4"],
                "Default Gateway": "1.2.3.1",
                "Gateway Metric": "0",
                "InterfaceMetric": "20",
                "Register with which suffix": "Primary only",
                "WINS servers configured through DHCP": ["None"],
                "ip_addrs": [
                    {
                        "IP Address": "1.2.3.74",
                        "Netmask": "255.255.255.0",
                        "Subnet": "1.2.3.0/24",
                    }
                ],
            }
        }

        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(win_ip.get_all_interfaces(), ret)

    # 'get_interface' function tests: 1

    def test_get_interface(self):
        """
        Test if it return the configuration of a network interface.
        """
        ret = {
            "DHCP enabled": "Yes",
            "DNS servers configured through DHCP": ["1.2.3.4"],
            "Default Gateway": "1.2.3.1",
            "Gateway Metric": "0",
            "InterfaceMetric": "20",
            "Register with which suffix": "Primary only",
            "WINS servers configured through DHCP": ["None"],
            "ip_addrs": [
                {
                    "IP Address": "1.2.3.74",
                    "Netmask": "255.255.255.0",
                    "Subnet": "1.2.3.0/24",
                }
            ],
        }

        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(win_ip.get_interface("Ethernet"), ret)

    # 'is_enabled' function tests: 1

    def test_is_enabled(self):
        """
        Test if it returns `True` if interface is enabled, otherwise `False`.
        """
        mock_cmd = MagicMock(side_effect=[ETHERNET_ENABLE, ""])
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertTrue(win_ip.is_enabled("Ethernet"))
            self.assertRaises(CommandExecutionError, win_ip.is_enabled, "Ethernet")

    # 'is_disabled' function tests: 1

    def test_is_disabled(self):
        """
        Test if it returns `True` if interface is disabled, otherwise `False`.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_ENABLE)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertFalse(win_ip.is_disabled("Ethernet"))

    # 'enable' function tests: 1

    def test_enable(self):
        """
        Test if it enable an interface.
        """
        # Test with enabled interface
        with patch.object(win_ip, "is_enabled", return_value=True):
            self.assertTrue(win_ip.enable("Ethernet"))

        mock_cmd = MagicMock()
        with patch.object(win_ip, "is_enabled", side_effect=[False, True]), patch.dict(
            win_ip.__salt__, {"cmd.run": mock_cmd}
        ):
            self.assertTrue(win_ip.enable("Ethernet"))

        mock_cmd.called_once_with(
            [
                "netsh",
                "interface",
                "set",
                "interface",
                "name=Ethernet",
                "admin=ENABLED",
            ],
            python_shell=False,
        )

    # 'disable' function tests: 1

    def test_disable(self):
        """
        Test if it disable an interface.
        """
        with patch.object(win_ip, "is_disabled", return_value=True):
            self.assertTrue(win_ip.disable("Ethernet"))

        mock_cmd = MagicMock()
        with patch.object(win_ip, "is_disabled", side_effect=[False, True]), patch.dict(
            win_ip.__salt__, {"cmd.run": mock_cmd}
        ):
            self.assertTrue(win_ip.disable("Ethernet"))

        mock_cmd.called_once_with(
            [
                "netsh",
                "interface",
                "set",
                "interface",
                "name=Ethernet",
                "admin=DISABLED",
            ],
            python_shell=False,
        )

    # 'get_subnet_length' function tests: 1

    def test_get_subnet_length(self):
        """
        Test if it disable an interface.
        """
        self.assertEqual(win_ip.get_subnet_length("255.255.255.0"), 24)
        self.assertRaises(SaltInvocationError, win_ip.get_subnet_length, "255.255.0")

    # 'set_static_ip' function tests: 1

    @pytest.mark.slow_test
    def test_set_static_ip(self):
        """
        Test if it set static IP configuration on a Windows NIC.
        """
        self.assertRaises(
            SaltInvocationError,
            win_ip.set_static_ip,
            "Local Area Connection",
            "10.1.2/24",
        )

        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        mock_all = MagicMock(return_value={"retcode": 1, "stderr": "Error"})
        with patch.dict(
            win_ip.__salt__, {"cmd.run": mock_cmd, "cmd.run_all": mock_all}
        ):
            self.assertRaises(
                CommandExecutionError,
                win_ip.set_static_ip,
                "Ethernet",
                "1.2.3.74/24",
                append=True,
            )
            self.assertRaises(
                CommandExecutionError, win_ip.set_static_ip, "Ethernet", "1.2.3.74/24"
            )

        mock_all = MagicMock(return_value={"retcode": 0})
        with patch.dict(
            win_ip.__salt__, {"cmd.run": mock_cmd, "cmd.run_all": mock_all}
        ):
            self.assertDictEqual(
                win_ip.set_static_ip("Local Area Connection", "1.2.3.74/24"), {}
            )
            self.assertDictEqual(
                win_ip.set_static_ip("Ethernet", "1.2.3.74/24"),
                {
                    "Address Info": {
                        "IP Address": "1.2.3.74",
                        "Netmask": "255.255.255.0",
                        "Subnet": "1.2.3.0/24",
                    }
                },
            )

    # 'set_dhcp_ip' function tests: 1

    def test_set_dhcp_ip(self):
        """
        Test if it set Windows NIC to get IP from DHCP.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(
                win_ip.set_dhcp_ip("Ethernet"),
                {"DHCP enabled": "Yes", "Interface": "Ethernet"},
            )

    # 'set_static_dns' function tests: 1

    def test_set_static_dns(self):
        """
        Test if it set static DNS configuration on a Windows NIC.
        """
        mock_cmd = MagicMock()
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(
                win_ip.set_static_dns("Ethernet", "192.168.1.252", "192.168.1.253"),
                {
                    "DNS Server": ("192.168.1.252", "192.168.1.253"),
                    "Interface": "Ethernet",
                },
            )
            mock_cmd.assert_has_calls(
                [
                    call(
                        [
                            "netsh",
                            "interface",
                            "ip",
                            "set",
                            "dns",
                            "name=Ethernet",
                            "source=static",
                            "address=192.168.1.252",
                            "register=primary",
                        ],
                        python_shell=False,
                    ),
                    call(
                        [
                            "netsh",
                            "interface",
                            "ip",
                            "add",
                            "dns",
                            "name=Ethernet",
                            "address=192.168.1.253",
                            "index=2",
                        ],
                        python_shell=False,
                    ),
                ]
            )

    def test_set_static_dns_clear(self):
        """
        Test if it set static DNS configuration on a Windows NIC.
        """
        mock_cmd = MagicMock()
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(
                win_ip.set_static_dns("Ethernet", []),
                {"DNS Server": [], "Interface": "Ethernet"},
            )
            mock_cmd.assert_called_once_with(
                [
                    "netsh",
                    "interface",
                    "ip",
                    "set",
                    "dns",
                    "name=Ethernet",
                    "source=static",
                    "address=none",
                ],
                python_shell=False,
            )

    def test_set_static_dns_no_action(self):
        """
        Test if it set static DNS configuration on a Windows NIC.
        """
        # Test passing nothing
        self.assertDictEqual(
            win_ip.set_static_dns("Ethernet"),
            {"DNS Server": "No Changes", "Interface": "Ethernet"},
        )
        # Test passing None
        self.assertDictEqual(
            win_ip.set_static_dns("Ethernet", None),
            {"DNS Server": "No Changes", "Interface": "Ethernet"},
        )

        # Test passing string None
        self.assertDictEqual(
            win_ip.set_static_dns("Ethernet", "None"),
            {"DNS Server": "No Changes", "Interface": "Ethernet"},
        )

    # 'set_dhcp_dns' function tests: 1

    def test_set_dhcp_dns(self):
        """
        Test if it set DNS source to DHCP on Windows.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(
                win_ip.set_dhcp_dns("Ethernet"),
                {"DNS Server": "DHCP", "Interface": "Ethernet"},
            )

    # 'set_dhcp_all' function tests: 1

    def test_set_dhcp_all(self):
        """
        Test if it set both IP Address and DNS to DHCP.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertDictEqual(
                win_ip.set_dhcp_all("Ethernet"),
                {"Interface": "Ethernet", "DNS Server": "DHCP", "DHCP enabled": "Yes"},
            )

    # 'get_default_gateway' function tests: 1

    def test_get_default_gateway(self):
        """
        Test if it set DNS source to DHCP on Windows.
        """
        mock_cmd = MagicMock(return_value=ETHERNET_CONFIG)
        with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(win_ip.get_default_gateway(), "1.2.3.1")



"""
A convenience system to manage reactors

Beginning in the 2017.7 release, the reactor runner requires that the reactor
system is running.  This is accomplished one of two ways, either
by having reactors configured or by including ``reactor`` in the
engine configuration for the Salt master.

    .. code-block:: yaml

    engines:
        - reactor

"""

import logging

import salt.config
import salt.syspaths
import salt.utils.event
import salt.utils.master
import salt.utils.process
import salt.utils.reactor
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

__func_alias__ = {
    "list_": "list",
}


def _reactor_system_available():
    """
    Helper to see if the reactor system is available
    """
    if __opts__.get("engines", {}):
        if any([True for engine in __opts__["engines"] if "reactor" in engine]):
            return True
    elif __opts__.get("reactor", {}) and __opts__["reactor"]:
        return True
    return False


def list_(saltenv="base", test=None):
    """
    List currently configured reactors

    CLI Example:

    .. code-block:: bash

        salt-run reactor.list
    """
    if not _reactor_system_available():
        raise CommandExecutionError("Reactor system is not running.")

    with salt.utils.event.get_event(
        "master",
        __opts__["sock_dir"],
        __opts__["transport"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        master_key = salt.utils.master.get_master_key("root", __opts__)

        __jid_event__.fire_event({"key": master_key}, "salt/reactors/manage/list")

        results = sevent.get_event(wait=30, tag="salt/reactors/manage/list-results")
        reactors = results.get("reactors")
        return reactors


def add(event, reactors, saltenv="base", test=None):
    """
    Add a new reactor

    CLI Example:

    .. code-block:: bash

        salt-run reactor.add 'salt/cloud/*/destroyed' reactors='/srv/reactor/destroy/*.sls'
    """
    if not _reactor_system_available():
        raise CommandExecutionError("Reactor system is not running.")

    if isinstance(reactors, str):
        reactors = [reactors]

    with salt.utils.event.get_event(
        "master",
        __opts__["sock_dir"],
        __opts__["transport"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        master_key = salt.utils.master.get_master_key("root", __opts__)

        __jid_event__.fire_event(
            {"event": event, "reactors": reactors, "key": master_key},
            "salt/reactors/manage/add",
        )

        res = sevent.get_event(wait=30, tag="salt/reactors/manage/add-complete")
        return res.get("result")


def delete(event, saltenv="base", test=None):
    """
    Delete a reactor

    CLI Example:

    .. code-block:: bash

        salt-run reactor.delete 'salt/cloud/*/destroyed'
    """
    if not _reactor_system_available():
        raise CommandExecutionError("Reactor system is not running.")

    with salt.utils.event.get_event(
        "master",
        __opts__["sock_dir"],
        __opts__["transport"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        master_key = salt.utils.master.get_master_key("root", __opts__)

        __jid_event__.fire_event(
            {"event": event, "key": master_key}, "salt/reactors/manage/delete"
        )

        res = sevent.get_event(wait=30, tag="salt/reactors/manage/delete-complete")
        return res.get("result")


def is_leader():
    """
    Return whether the running reactor is acting as a leader (responding to events).

    CLI Example:

    .. code-block:: bash

        salt-run reactor.is_leader
    """
    if not _reactor_system_available():
        raise CommandExecutionError("Reactor system is not running.")

    with salt.utils.event.get_event(
        "master",
        __opts__["sock_dir"],
        __opts__["transport"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        master_key = salt.utils.master.get_master_key("root", __opts__)

        __jid_event__.fire_event({"key": master_key}, "salt/reactors/manage/is_leader")

        res = sevent.get_event(wait=30, tag="salt/reactors/manage/leader/value")
        return res["result"]


def set_leader(value=True):
    """
    Set the current reactor to act as a leader (responding to events). Defaults to True

    CLI Example:

    .. code-block:: bash

        salt-run reactor.set_leader True
    """
    if not _reactor_system_available():
        raise CommandExecutionError("Reactor system is not running.")

    with salt.utils.event.get_event(
        "master",
        __opts__["sock_dir"],
        __opts__["transport"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        master_key = salt.utils.master.get_master_key("root", __opts__)

        __jid_event__.fire_event(
            {"id": __opts__["id"], "value": value, "key": master_key},
            "salt/reactors/manage/set_leader",
        )

        res = sevent.get_event(wait=30, tag="salt/reactors/manage/leader/value")
        return res["result"]



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import os

import salt.states.virtualenv_mod as virtualenv_mod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class VirtualenvModTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the virtualenv_mod state
    """

    def setup_loader_modules(self):
        return {virtualenv_mod: {"__env__": "base"}}

    def test_managed(self):
        """
        Test to create a virtualenv and optionally manage it with pip
        """
        ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
        ret.update({"comment": "Virtualenv was not detected on this system"})
        self.assertDictEqual(virtualenv_mod.managed("salt"), ret)

        mock1 = MagicMock(return_value="True")
        mock = MagicMock(return_value=False)
        mock2 = MagicMock(return_value="1.1")
        with patch.dict(
            virtualenv_mod.__salt__,
            {
                "virtualenv.create": True,
                "cp.is_cached": mock,
                "cp.cache_file": mock,
                "cp.hash_file": mock,
                "pip.freeze": mock1,
                "cmd.run_stderr": mock1,
                "pip.version": mock2,
            },
        ):
            mock = MagicMock(side_effect=[True, True, True, False, True, True])
            with patch.object(os.path, "exists", mock):
                ret.update({"comment": "pip requirements file 'salt://a' not found"})
                self.assertDictEqual(
                    virtualenv_mod.managed("salt", None, "salt://a"), ret
                )

                with patch.dict(virtualenv_mod.__opts__, {"test": True}):
                    ret.update(
                        {
                            "changes": {"cleared_packages": "True", "old": "True"},
                            "comment": "Virtualenv salt is set to be cleared",
                            "result": None,
                        }
                    )
                    self.assertDictEqual(virtualenv_mod.managed("salt", clear=1), ret)
                    ret.update(
                        {
                            "comment": "Virtualenv salt is already created",
                            "changes": {},
                            "result": True,
                        }
                    )
                    self.assertDictEqual(virtualenv_mod.managed("salt"), ret)

                    ret.update(
                        {
                            "comment": "Virtualenv salt is set to be created",
                            "result": None,
                        }
                    )
                    self.assertDictEqual(virtualenv_mod.managed("salt"), ret)

                with patch.dict(virtualenv_mod.__opts__, {"test": False}):
                    ret.update(
                        {
                            "comment": (
                                "The 'use_wheel' option is"
                                " only supported in pip between 1.4 and 9.0.3."
                                " The version of pip detected was 1.1."
                            ),
                            "result": False,
                        }
                    )
                    self.assertDictEqual(
                        virtualenv_mod.managed("salt", use_wheel=1), ret
                    )

                    ret.update({"comment": "virtualenv exists", "result": True})
                    self.assertDictEqual(virtualenv_mod.managed("salt"), ret)



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(scope="module")
def state_tree(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "map.jinja" import abc with context %}

    Ok with {{ abc }}:
      test.succeed_without_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )

    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.mark.slow_test
def test_state_with_import(salt_ssh_cli, state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run("state.sls", "test")
    assert ret.exitcode == 0
    assert ret.json


@pytest.fixture
def nested_state_tree(base_env_state_tree_root_dir, tmpdir):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    state_file = """
    /{}/file.txt:
      file.managed:
        - source: salt://foo/file.jinja
        - template: jinja
    """.format(
        tmpdir
    )
    file_jinja = """
    {% from 'foo/map.jinja' import comment %}{{ comment }}
    """
    map_file = """
    {% set comment = "blah blah" %}
    """
    statedir = base_env_state_tree_root_dir / "foo"
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file("map.jinja", map_file, statedir)
    file_tempfile = pytest.helpers.temp_file("file.jinja", file_jinja, statedir)
    state_tempfile = pytest.helpers.temp_file("init.sls", state_file, statedir)

    with top_tempfile, map_tempfile, state_tempfile, file_tempfile:
        yield


@pytest.mark.slow_test
def test_state_with_import_from_dir(salt_ssh_cli, nested_state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run(
        "--extra-filerefs=salt://foo/map.jinja", "state.apply", "foo"
    )
    assert ret.exitcode == 0
    assert ret.json
